# distutils: language = c++

cimport cython
from libc.stdlib cimport malloc, free
import numpy as np
cimport numpy as np
import pycalphad.variables as v
import ctypes

cdef class FastFunction:
    """``FastFunction`` provides a stable(-ish) interface that encapsulates SymEngine function pointers.
    """
    def __cinit__(self, object func):
        if func is None:
            self.f_ptr = NULL
            self.func_data = NULL
            return
        if isinstance(func, FastFunction):
            func = func._objref
        # Preserve reference to object to prevent garbage collection
        self._objref = func
        addr1, addr2 = func.as_ctypes()
        self.f_ptr = (<math_function_t*><size_t>ctypes.addressof(addr1))[0]
        self.func_data =  (<void**><size_t>ctypes.addressof(addr2))[0]
    def __reduce__(self):
        return FastFunction, (self._objref,)
    cdef void call(self, double *out, double *inp) nogil:
        if self.f_ptr != NULL:
            self.f_ptr(out, inp, self.func_data)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef double* alloc_dof_with_parameters(double[::1] dof, double[::1] parameters) nogil:
    """Remember to free() if parameters.shape[0] > 0"""
    cdef double* dof_concat
    cdef int j
    cdef int num_dof = dof.shape[0] + parameters.shape[0]
    if parameters.shape[0] == 0:
        dof_concat = &dof[0]
    else:
        dof_concat = <double *> malloc(num_dof * sizeof(double))
        for j in range(0,dof.shape[0]):
            dof_concat[j] = dof[j]
        for j in range(dof.shape[0], num_dof):
            dof_concat[j] = parameters[j - dof.shape[0]]
    return dof_concat

@cython.boundscheck(False)
@cython.wraparound(False)
cdef double* alloc_dof_with_parameters_vectorized(double[:, ::1] dof, double[::1] parameters) nogil:
    """Remember to free() if parameters.shape[0] > 0"""
    cdef double* dof_concat
    cdef int i, j
    cdef int num_inps = dof.shape[0]
    cdef int num_dof = dof.shape[1] + parameters.shape[0]

    if parameters.shape[0] == 0:
        dof_concat = &dof[0, 0]
    else:
        dof_concat = <double *> malloc(num_inps * num_dof * sizeof(double))
        for i in range(num_inps):
            for j in range(0,dof.shape[1]):
                dof_concat[i * num_dof + j] = dof[i, j]
            for j in range(dof.shape[1], num_dof):
                dof_concat[i * num_dof + j] = parameters[j - dof.shape[1]]
    return dof_concat


cdef public class PhaseRecord(object)[type PhaseRecordType, object PhaseRecordObject]:
    """
    This object exposes a common API to the solver so it doesn't need to know about the differences
    between Model implementations. PhaseRecords are immutable after initialization.
    """
    def __reduce__(self):
            return PhaseRecord, (self.components, self.state_variables, self.variables, np.array(self.parameters),
                                 self.ofunc_,
                                 self.formulaofunc_, self.formulagfunc_, self.formulahfunc_,
                                 self.massfuncs_,
                                 self.formulamolefuncs_, self.formulamolegradfuncs_, self.formulamolehessianfuncs_,
                                 self.internal_cons_func_, self.internal_cons_jac_, self.internal_cons_hess_,
                                 self.num_internal_cons)

    def __cinit__(self, object comps, object state_variables, object variables,
                  double[::1] parameters, object ofunc,
                  object formulaofunc, object formulagfunc, object formulahfunc,
                  object massfuncs,
                  object formulamolefuncs, object formulamolegradfuncs, object formulamolehessianfuncs,
                  object internal_cons_func, object internal_cons_jac, object internal_cons_hess,
                  size_t num_internal_cons):
        cdef:
            int var_idx, el_idx
        self.components = comps
        desired_active_pure_elements = [list(x.constituents.keys()) for x in self.components]
        desired_active_pure_elements = [el.upper() for constituents in desired_active_pure_elements for el in constituents]
        pure_elements = sorted(set(desired_active_pure_elements))
        nonvacant_elements = sorted([x for x in set(desired_active_pure_elements) if x != 'VA'])

        self.variables = variables
        self.state_variables = state_variables
        self.num_statevars = len(state_variables)
        self.pure_elements = pure_elements
        self.nonvacant_elements = nonvacant_elements
        self.phase_dof = 0
        self.parameters = parameters
        self.num_internal_cons = num_internal_cons

        for variable in variables:
            if not isinstance(variable, v.SiteFraction):
                continue
            self.phase_name = <unicode>variable.phase_name
            self.phase_dof += 1

        # Used only to reconstitute if pickled (i.e. via __reduce__)
        self.ofunc_ = ofunc
        self.formulaofunc_ = formulaofunc
        self.formulagfunc_ = formulagfunc
        self.formulahfunc_ = formulahfunc
        self.internal_cons_func_ = internal_cons_func
        self.internal_cons_jac_ = internal_cons_jac
        self.internal_cons_hess_ = internal_cons_hess
        self.massfuncs_ = massfuncs
        self.formulamolefuncs_ = formulamolefuncs
        self.formulamolegradfuncs_ = formulamolegradfuncs
        self.formulamolehessianfuncs_ = formulamolehessianfuncs

        if ofunc is not None:
            self._obj = FastFunction(ofunc)
        if formulaofunc is not None:
            self._formulaobj = FastFunction(formulaofunc)
        if formulagfunc is not None:
            self._formulagrad = FastFunction(formulagfunc)
        if formulahfunc is not None:
            self._formulahess = FastFunction(formulahfunc)
        if internal_cons_func is not None:
            self._internal_cons_func = FastFunction(internal_cons_func)
        if internal_cons_jac is not None:
            self._internal_cons_jac = FastFunction(internal_cons_jac)
        if internal_cons_hess is not None:
            self._internal_cons_hess = FastFunction(internal_cons_hess)
        if massfuncs is not None:
            self._masses = np.empty(len(nonvacant_elements), dtype='object')
            for el_idx in range(len(nonvacant_elements)):
                self._masses[el_idx] = FastFunction(massfuncs[el_idx])
            self._masses_ptr = <void**> self._masses.data
        if formulamolefuncs is not None:
            self._formulamoles = np.empty(len(nonvacant_elements), dtype='object')
            for el_idx in range(len(nonvacant_elements)):
                self._formulamoles[el_idx] = FastFunction(formulamolefuncs[el_idx])
            self._formulamoles_ptr = <void**> self._formulamoles.data
        if formulamolegradfuncs is not None:
            self._formulamolegrads = np.empty(len(nonvacant_elements), dtype='object')
            for el_idx in range(len(nonvacant_elements)):
                self._formulamolegrads[el_idx] = FastFunction(formulamolegradfuncs[el_idx])
            self._formulamolegrads_ptr = <void**> self._formulamolegrads.data
        if formulamolehessianfuncs is not None:
            self._formulamolehessians = np.empty(len(nonvacant_elements), dtype='object')
            for el_idx in range(len(nonvacant_elements)):
                self._formulamolehessians[el_idx] = FastFunction(formulamolehessianfuncs[el_idx])
            self._formulamolehessians_ptr = <void**> self._formulamolehessians.data

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void obj(self, double[::1] outp, double[::1] dof) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        cdef int num_dof = self.num_statevars + self.phase_dof + self.parameters.shape[0]
        self._obj.call(&outp[0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void formulaobj(self, double[::1] outp, double[::1] dof) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        cdef int num_dof = self.num_statevars + self.phase_dof + self.parameters.shape[0]
        self._formulaobj.call(&outp[0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void obj_2d(self, double[::1] outp, double[:, ::1] dof) nogil:
        # dof.shape[1] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters_vectorized(dof[:, :self.num_statevars+self.phase_dof], self.parameters)
        cdef int i
        cdef int num_inps = dof.shape[0]
        cdef int num_dof = self.num_statevars + self.phase_dof + self.parameters.shape[0]
        for i in range(num_inps):
            self._obj.call(&outp[i], &dof_concat[i * num_dof])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void obj_parameters_2d(self, double[:, ::1] outp, double[:, ::1] dof, double[:, ::1] parameters) nogil:
        """
        Calculate objective function using custom parameters.
        Note dof and parameters are vectorized separately, i.e., broadcast against each other.
        Let dof.shape[0] = M and parameters.shape[0] = N
        Then outp.shape = (M,N)
        """
        # dof.shape[1] may be oversized by the caller; do not trust it
        cdef size_t i, j, dof_idx, param_idx
        cdef size_t num_dof_inps = dof.shape[0]
        cdef size_t num_param_inps = parameters.shape[0]
        # We are trusting parameters.shape[1] to be sized correctly here
        cdef size_t num_params = parameters.shape[1]
        cdef size_t num_dof = self.num_statevars + self.phase_dof + num_params
        cdef size_t dof_offset = self.num_statevars + self.phase_dof
        cdef double* dof_concat = <double *> malloc(num_param_inps * num_dof * sizeof(double))
        for i in range(num_dof_inps):
            # Initialize all parameter arrays with current dof
            for j in range(num_param_inps):
                for dof_idx in range(num_dof-num_params):
                    dof_concat[j * num_dof + dof_idx] = dof[i, dof_idx]
                for param_idx in range(num_params):
                    dof_concat[j * num_dof + dof_offset + param_idx] = parameters[j, param_idx]
            for j in range(num_param_inps):
                self._obj.call(&outp[i,j], &dof_concat[j * num_dof])
        free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void formulagrad(self, double[::1] out, double[::1] dof) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        self._formulagrad.call(&out[0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)


    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void formulahess(self, double[:, ::1] out, double[::1] dof) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        self._formulahess.call(&out[0,0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)


    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void internal_cons_func(self, double[::1] out, double[::1] dof) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        self._internal_cons_func.call(&out[0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void internal_cons_jac(self, double[:, ::1] out, double[::1] dof) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        self._internal_cons_jac.call(&out[0, 0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void internal_cons_hess(self, double[:, :, ::1] out, double[::1] dof) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        self._internal_cons_hess.call(&out[0, 0, 0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void mass_obj(self, double[::1] out, double[::1] dof, int comp_idx) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        (<FastFunction>self._masses_ptr[comp_idx]).call(&out[0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void mass_obj_2d(self, double[::1] out, double[:, ::1] dof, int comp_idx) nogil:
        # dof.shape[1] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters_vectorized(dof[:, :self.num_statevars+self.phase_dof], self.parameters)
        cdef int i
        cdef int num_inps = dof.shape[0]
        cdef int num_dof = self.num_statevars + self.phase_dof + self.parameters.shape[0]
        for i in range(num_inps):
            (<FastFunction>self._masses_ptr[comp_idx]).call(&out[i], &dof_concat[i * num_dof])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void formulamole_obj(self, double[::1] out, double[::1] dof, int comp_idx) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        (<FastFunction>self._formulamoles_ptr[comp_idx]).call(&out[0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void formulamole_grad(self, double[::1] out, double[::1] dof, int comp_idx) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        (<FastFunction>self._formulamolegrads_ptr[comp_idx]).call(&out[0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void formulamole_hess(self, double[:,::1] out, double[::1] dof, int comp_idx) nogil:
        # dof.shape[0] may be oversized by the caller; do not trust it
        cdef double* dof_concat = alloc_dof_with_parameters(dof[:self.num_statevars+self.phase_dof], self.parameters)
        (<FastFunction>self._formulamolehessians_ptr[comp_idx]).call(&out[0,0], &dof_concat[0])
        if self.parameters.shape[0] > 0:
            free(dof_concat)




"""
The database module provides support for reading and writing data types
associated with structured thermodynamic/kinetic data.
"""
from io import StringIO
from tinydb import TinyDB
from tinydb.storages import MemoryStorage
from datetime import datetime
from collections import namedtuple
import os
from pycalphad.variables import Species
from pycalphad.core.cache import fhash
from pycalphad.core.utils import recursive_tuplify


class DatabaseExportError(Exception):
    """Raised when a database cannot be written."""
    pass


class Phase(object): #pylint: disable=R0903
    """
    Phase in the database.

    Attributes
    ----------
    name : string
        System-local name of the phase.
    constituents : tuple of frozenset
        Possible sublattice constituents (elements and/or species).
    sublattices : list
        Site ratios of sublattices.
    model_hints : dict
        Structured "hints" for a Model trying to read this phase.
        Hints for major constituents and typedefs (Thermo-Calc) go here.
    """
    def __init__(self):
        self.name = None
        self.constituents = None
        self.sublattices = []
        self.model_hints = {}
    def __eq__(self, other):
        if type(self) == type(other):
            return self.__dict__ == other.__dict__
        else:
            return False
    def __ne__(self, other):
        return not self.__eq__(other)
    def __repr__(self):
        return 'Phase({0!r})'.format(self.__dict__)
    def __hash__(self):
        return hash((self.name, self.constituents, tuple(self.sublattices),
                     tuple(sorted(recursive_tuplify(self.model_hints.items())))))

DatabaseFormat = namedtuple('DatabaseFormat', ['read', 'write'])
format_registry = {}


class Database(object): #pylint: disable=R0902
    """
    Structured thermodynamic and/or kinetic data.

    Attributes
    ----------
    elements : set
        Set of elements in database.
    species : set
        Set of species in database.
    phases : dict
        Phase objects indexed by their system-local name.
    symbols : dict
        SymEngine objects indexed by their name (FUNCTIONs in Thermo-Calc).
    references : dict
        Reference objects indexed by their system-local identifier.

    Examples
    --------
    >>> mydb = Database(open('crfeni_mie.tdb'))
    >>> mydb = Database('crfeni_mie.tdb')
    >>> f = StringIO(u'$a complete TDB file as a string\\n')
    >>> mydb = Database(f)
    """
    def __new__(cls, *args):
        if len(args) == 0:
            obj = super(Database, cls).__new__(cls, *args)
            # Should elements be rolled into a special case of species?
            obj.elements = set()
            obj.species = set()
            obj.phases = {}
            obj.refstates = {}
            obj._structure_dict = {} # System-local phase names to global IDs
            obj._parameters = TinyDB(storage=MemoryStorage)
            obj._parameter_queue = []
            obj.symbols = {}
            obj.references = {}
            # Note: No public typedefs here (from TDB files)
            # Instead we put that information in the model_hint for phases
            return obj
        elif len(args) == 1:
            fname = args[0]
            # Backwards compatibility: assume TDB by default
            fmt = 'tdb'
            # Attempt to auto-detect the correct format based on the file extension
            try:
                path, ext = os.path.splitext(fname)
                if '.' in ext and ext[1:].lower() in format_registry:
                    fmt = ext[1:].lower()
            except (AttributeError, TypeError):
                pass
            if hasattr(fname, 'read'):
                # File descriptor
                return cls.from_file(fname, fmt=fmt)
            elif fname.find('\n') == -1:
                # Single-line string; it's probably a filename
                return cls.from_file(fname, fmt=fmt)
            else:
                # Newlines found: probably a full database string
                return cls.from_string(fname, fmt=fmt)
        else:
            raise ValueError('Invalid number of parameters: '+len(args))

    def __hash__(self):
        return fhash(self.__dict__)


    def __getstate__(self):
        pickle_dict = {}
        for key, value in self.__dict__.items():
            if key == '_parameters':
                pickle_dict[key] = value.all()
            else:
                pickle_dict[key] = value
        return pickle_dict

    def __setstate__(self, state):
        for key, value in state.items():
            if key == '_parameters':
                self._parameters = TinyDB(storage=MemoryStorage)
                self._parameters.insert_multiple(value)
            else:
                setattr(self, key, value)

    def __deepcopy__(self, memo):
        copy = type(self)()
        memo[id(self)] = copy
        for key, value in self.__dict__.items():
            if key == '_parameters':
                copy._parameters = TinyDB(storage=MemoryStorage)
                copy._parameters.insert_multiple(value.all())
            else:
                setattr(copy, key, value)
        return copy

    @staticmethod
    def register_format(fmt, read=None, write=None):
        """
        Add support for reading and/or writing the specified format.

        Parameters
        ----------
        fmt: str
            Format.
        read : callable, optional
            Read function with arguments (Database, file_descriptor)
        write : callable, optional
            Write function with arguments (Database, file_descriptor)

        Examples
        --------
        None yet.
        """
        format_registry[fmt.lower()] = DatabaseFormat(read=read, write=write)

    @staticmethod
    def from_file(fname, fmt=None):
        """
        Create a Database from a file.

        Parameters
        ----------
        fname: str or file-like
            File name/descriptor to read.
        fmt : str, optional
            File format. If not specified, an attempt at auto-detection is made.

        Returns
        -------
        dbf : Database
            Database from file.

        Examples
        --------
        None yet.
        """
        if fmt is None:
            # Attempt to auto-detect the correct format based on the file extension
            try:
                path, ext = os.path.splitext(fname)
            except (AttributeError, TypeError):
                # fname isn't actually a path, so we don't know the correct format
                raise ValueError('\'fmt\' keyword argument must be specified when passing a file descriptor.')
            if '.' in ext and ext[1:].lower() in format_registry:
                fmt = ext[1:].lower()
        else:
            fmt = fmt.lower()
        if fmt not in format_registry or format_registry[fmt].read is None:
            supported_reads = [key for key, value in format_registry.items() if value.read is not None]
            raise NotImplementedError('Unsupported read format \'{0}\'. Supported formats: {1}'.format(fmt,
                                                                                                        supported_reads))
        # Is it a file descriptor?
        if hasattr(fname, 'read'):
            fd = fname
            need_to_close = False
        else:
            # It's not file-like, so it's probably a filename
            need_to_close = True
            fd = open(fname, mode='r')
        try:
            dbf = Database()
            format_registry[fmt.lower()].read(dbf, fd)
        finally:
            # Close file descriptors created in this routine
            # Otherwise that's left up to the calling function
            if need_to_close:
                fd.close()

        return dbf

    @classmethod
    def from_string(cls, data, **kwargs):
        """
        Returns Database from a string in the specified format.
        This function is a wrapper for calling `from_file` with StringIO.

        Parameters
        ----------
        data : str
            Raw database string in the specified format.
        kwargs : optional
            See keyword arguments for `from_file`.

        Returns
        -------
        dbf : Database
        """
        return cls.from_file(StringIO(data), **kwargs)

    def to_file(self, fname, fmt=None, if_exists='raise', **write_kwargs):
        """
        Write the Database to a file.

        Parameters
        ----------
        fname: str or file-like
            File name/descriptor to write.
        fmt : str, optional
            File format. If not specified, an attempt at auto-detection is made.
        if_exists : string, optional ['raise', 'rename', 'overwrite']
            Strategy if 'fname' already exists.
            The 'raise' option (default) will raise a FileExistsError.
            The 'rename' option will append the date/time to the filename.
            The 'overwrite' option will overwrite the file.
            This argument is ignored if 'fname' is file-like.
        write_kwargs : optional
            Keyword arguments to pass to write function.

        Examples
        --------
        None yet.
        """
        if fmt is None:
            # Attempt to auto-detect the correct format based on the file extension
            try:
                path, ext = os.path.splitext(fname)
            except (AttributeError, TypeError):
                # fname isn't actually a path, so we don't know the correct format
                raise ValueError('\'fmt\' keyword argument must be specified when passing a file descriptor.')
            if '.' in ext and ext[1:].lower() in format_registry:
                fmt = ext[1:].lower()
        else:
            fmt = fmt.lower()
        if fmt not in format_registry or format_registry[fmt].write is None:
            supported_writes = [key for key, value in format_registry.items() if value.write is not None]
            raise NotImplementedError('Unsupported write format \'{0}\'. Supported formats: {1}'.format(fmt,
                                                                                                        supported_writes))
        # Is this a file descriptor?
        if hasattr(fname, 'write'):
            format_registry[fmt].write(self, fname, **write_kwargs)
        else:
            if os.path.exists(fname) and if_exists != 'overwrite':
                if if_exists == 'rename':
                    writetime = datetime.now()
                    fname = os.path.splitext(fname)
                    fname = fname[0] + "." + writetime.strftime("%Y-%m-%d-%H-%M") + fname[1]
                else:
                    # equivalent to 'raise'
                    raise FileExistsError('File {} already exists'.format(fname))
            with open(fname, mode='w') as fd:
                format_registry[fmt].write(self, fd, **write_kwargs)

    def to_string(self, **kwargs):
        """
        Returns Database as a string.
        This function is a wrapper for calling `to_file` with StringIO.

        Parameters
        ----------
        kwargs : optional
            See keyword arguments for `to_file`.

        Returns
        -------
        result : str
        """
        result = StringIO()
        self.to_file(result, **kwargs)
        return result.getvalue()

    def __str__(self):
        result = 'Elements: {0}\n'.format(sorted(self.elements))
        result += 'Species: {0}\n'.format(sorted(self.species, key=lambda s: s.name))
        for name, phase in sorted(self.phases.items()):
            result += str(phase)+'\n'
        result += '{0} symbols in database\n'.format(len(self.symbols))
        result += '{0} parameters in database\n'.format(len(self._parameters))
        return result

    def __eq__(self, other):
        if self is other:
            return True
        elif type(self) != type(other):
            return False
        elif sorted(self.__dict__.keys()) != sorted(other.__dict__.keys()):
            return False
        else:
            def param_sort_key(x):
                return x['phase_name'], x['parameter_type'], x['constituent_array'], \
                       x['parameter_order'], x['diffusing_species']
            for key in self.__dict__.keys():
                if key == '_parameters':
                    # Special handling for TinyDB objects
                    if len(self._parameters.all()) != len(other._parameters.all()):
                        return False
                    self_params = sorted(self._parameters.all(), key=param_sort_key)
                    other_params = sorted(other._parameters.all(), key=param_sort_key)
                    if self_params != other_params:
                        return False
                elif self.__dict__[key] != other.__dict__[key]:
                    return False
            return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def add_structure_entry(self, local_name, global_name):
        """
        Define a relation between the system-local name of a phase and a
        "global" identifier. This is used to link crystallographically
        similar phases known by different colloquial names.

        Parameters
        ----------
        local_name : string
            System-local name of the phase.
        global_name : object
            Abstract representation of symbol, e.g., in SymEngine format.

        Examples
        --------
        None yet.
        """
        self._structure_dict[local_name] = global_name

    def add_parameter(
        self, param_type, phase_name, constituent_array, param_order, param, ref=None,
        diffusing_species=None, force_insert=True, **kwargs,
        ):
        """
        Add a parameter.

        Parameters
        ----------
        param_type : str
            Type name of the parameter, e.g., G, L, BMAGN.
        phase_name : str
            Name of the phase.
        constituent_array : list
            Configuration of the sublattices (elements and/or species).
        param_order : int
            Polynomial order of the parameter.
        param : object
            Abstract representation of the parameter, e.g., in SymEngine format.
        ref : str, optional
            Reference for the parameter.
        diffusing_species : str, optional
            (If kinetic parameter) Diffusing species for this parameter.
        force_insert : bool, optional
            If True, inserts into the database immediately. False is a delayed insert (for performance).
        kwargs : Any
            Additional metadata to insert into the parameter dictionary

        Examples
        --------
        None yet.
        """
        species_dict = {s.name: s for s in self.species}
        new_parameter = {
            'phase_name': phase_name,
            'constituent_array': tuple(tuple(species_dict.get(s.upper(), Species(s)) for s in xs) for xs in constituent_array),  # must be hashable type
            'parameter_type': param_type,
            'parameter_order': param_order,
            'parameter': param,
            'diffusing_species': Species(diffusing_species),
            'reference': ref
        }
        new_parameter.update(kwargs)
        if force_insert:
            self._parameters.insert(new_parameter)
        else:
            self._parameter_queue.append(new_parameter)

    def add_phase(self, phase_name, model_hints, sublattices):
        """
        Add a phase.

        Parameters
        ----------
        phase_name : string
            System-local name of the phase.
        model_hints : dict
            Structured "hints" for a Model trying to read this phase.
            Hints for major constituents and typedefs (Thermo-Calc) go here.
        sublattices : list
            Site ratios of sublattices.

        Examples
        --------
        None yet.
        """
        new_phase = Phase()
        new_phase.name = phase_name
        # Need to convert from ParseResults or else equality testing will break
        new_phase.sublattices = tuple(sublattices)
        new_phase.model_hints = model_hints
        self.phases[phase_name] = new_phase
    def add_phase_constituents(self, phase_name, constituents):
        """
        Add a phase.

        Parameters
        ----------
        phase_name : string
            System-local name of the phase.
        constituents : list
            Possible phase constituents (elements and/or species).

        Examples
        --------
        None yet.
        """
        species_dict = {s.name: s for s in self.species}
        try:
            # Need to convert constituents from ParseResults
            # Otherwise equality testing will be broken
            self.phases[phase_name].constituents = tuple([frozenset([species_dict[s.upper()] for s in xs]) for xs in constituents])
        except KeyError:
            print("Undefined phase "+phase_name)
            raise
    def search(self, query):
        """
        Search for parameters matching the specified query.

        Parameters
        ----------
        query : object
            Structured database query in TinyDB format.

        Examples
        --------
        >>>> from tinydb import where
        >>>> db = Database()
        >>>> eid = db.add_parameter(...) #TODO
        >>>> db.search(where('eid') == eid)
        """
        return self._parameters.search(query)

    def process_parameter_queue(self):
        """
        Process the queue of parameters so they are added to the TinyDB in one transaction.
        This avoids repeated (expensive) calls to insert().
        """
        result = self._parameters.insert_multiple(self._parameter_queue)
        self._parameter_queue = []
        return result



class CalculateError(Exception):
    "Exception related to use of calculate() function."


class DofError(Exception):
    "Error due to missing degrees of freedom."
    pass


class EquilibriumError(Exception):
    "Exception related to calculation of equilibrium."
    pass


class ConditionError(CalculateError, EquilibriumError):
    "Exception related to calculation conditions."
    pass



# distutils: language = c++
from pycalphad.core.phase_rec cimport PhaseRecord
cimport numpy as np
import numpy as np
from libc.string cimport memset
cimport cython

cdef public class CompositionSet(object)[type CompositionSetType, object CompositionSetObject]:
    """
    This is the primary object the solver interacts with. It keeps the state of a phase (P, T, y...) at a
    particular solver iteration and can be updated using the update() member function. Every CompositionSet
    has a reference to a particular PhaseRecord which describes the prototype of the phase. These objects
    can be created and destroyed by the solver as needed to describe the stable set of phases. Multiple
    CompositionSets can point to the same PhaseRecord for the case of miscibility gaps. CompositionSets are
    not pickleable. They are used in miscibility gap deteciton.
    """
    def __cinit__(self, PhaseRecord prx):
        self.phase_record = prx
        self.dof = np.zeros(len(self.phase_record.variables)+len(self.phase_record.state_variables))
        self.X = np.zeros(len(self.phase_record.nonvacant_elements))
        self._X_2d_view = <double[:self.X.shape[0],:1]>&self.X[0]
        self.energy = 0
        self.NP = 0
        self.fixed = False
        self._energy_2d_view = <double[:1]>&self.energy

    def __deepcopy__(self, memodict=None):
        cdef CompositionSet other
        memodict = {} if memodict is None else memodict
        other = CompositionSet(self.phase_record)
        other.phase_record = self.phase_record
        other.dof[:] = self.dof
        other.X[:] = self.X
        other.energy = 1.0*self.energy
        other._energy_2d_view = <double[:1]>&other.energy
        other.NP = 1.0*self.NP
        other.fixed = bool(self.fixed)
        return other

    def __repr__(self):
        return str(self.__class__.__name__) + "({0}, {1}, NP={2}, GM={3})".format(self.phase_record.phase_name,
                                                                          np.asarray(self.X), self.NP, self.energy)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void update(self, double[::1] site_fracs, double phase_amt, double[::1] state_variables):
        cdef int comp_idx
        self.dof[:state_variables.shape[0]] = state_variables
        self.dof[state_variables.shape[0]:] = site_fracs
        self.NP = phase_amt
        self.energy = 0
        memset(&self.X[0], 0, self.X.shape[0] * sizeof(double))
        self.phase_record.obj(self._energy_2d_view, self.dof)
        for comp_idx in range(self.X.shape[0]):
            self.phase_record.mass_obj(self._X_2d_view[comp_idx], self.dof, comp_idx)



"""
The lower_convex_hull module handles geometric calculations associated with
equilibrium calculation.
"""
from pycalphad.core.cartesian import cartesian
from pycalphad.core.constants import MIN_SITE_FRACTION
from .hyperplane import hyperplane
import numpy as np
import itertools


def lower_convex_hull(global_grid, state_variables, result_array):
    """
    Find the simplices on the lower convex hull satisfying the specified
    conditions in the result array.

    Parameters
    ----------
    global_grid : Dataset
        A sample of the energy surface of the system.
    state_variables : List[v.StateVariable]
        A list of the state variables (e.g., P, T) used in this calculation.
    result_array : Dataset
        This object will be modified!
        Coordinates correspond to conditions axes.

    Returns
    -------
    None. Results are written to result_array.

    Notes
    -----
    This routine will not check if any simplex is degenerate.
    Degenerate simplices will manifest with duplicate or NaN indices.

    Examples
    --------
    None yet.
    """
    state_variables = sorted(state_variables, key=str)
    comp_conds = sorted([x for x in sorted(result_array.coords.keys()) if x.startswith('X_')])
    comp_conds_indices = sorted([idx for idx, x in enumerate(sorted(result_array.coords['component']))
                                 if 'X_'+x in comp_conds])
    comp_conds_indices = np.array(comp_conds_indices, dtype=np.uintp)
    pot_conds = sorted([x for x in sorted(result_array.coords.keys()) if x.startswith('MU_')])
    pot_conds_indices = sorted([idx for idx, x in enumerate(sorted(result_array.coords['component']))
                                if 'MU_'+x in pot_conds])
    pot_conds_indices = np.array(pot_conds_indices, dtype=np.uintp)

    if len(set(pot_conds_indices) & set(comp_conds_indices)) > 0:
        raise ValueError('Cannot specify component chemical potential and amount simultaneously')

    if len(comp_conds) > 0:
        cart_values = cartesian([result_array.coords[cond] for cond in comp_conds])
    else:
        cart_values = np.atleast_2d(1.)
    # TODO: Handle W(comp) as well as X(comp) here
    comp_values = np.zeros(cart_values.shape[:-1] + (len(result_array.coords['component']),))
    for idx in range(comp_values.shape[-1]):
        if idx in comp_conds_indices:
            comp_values[..., idx] = cart_values[..., np.where(comp_conds_indices == idx)[0][0]]
        elif idx in pot_conds_indices:
            # Composition value not used
            comp_values[..., idx] = 0
        else:
            # Dependent component (composition value not used)
            comp_values[..., idx] = 0
    # Prevent compositions near an edge from going negative
    comp_values[np.nonzero(comp_values < MIN_SITE_FRACTION)] = MIN_SITE_FRACTION*10

    if len(pot_conds) > 0:
        cart_pot_values = cartesian([result_array.coords[cond] for cond in pot_conds])

    #result_array['Phase'] = force_indep_align(result_array.Phase)
    # factored out via profiling
    result_array_GM_values = result_array.GM
    result_array_GM_dims = result_array.data_vars['GM'][0]
    result_array_points_values = result_array.points
    result_array_MU_values = result_array.MU
    result_array_NP_values = result_array.NP
    result_array_X_values = result_array.X
    result_array_Y_values = result_array.Y
    result_array_Phase_values = result_array.Phase
    global_grid_GM_values = global_grid.GM
    global_grid_X_values = global_grid.X
    global_grid_Y_values = global_grid.Y
    global_grid_Phase_values = global_grid.Phase
    num_comps = len(result_array.coords['component'])

    it = np.nditer(result_array_GM_values, flags=['multi_index'])
    comp_coord_shape = tuple(len(result_array.coords[cond]) for cond in comp_conds)
    pot_coord_shape = tuple(len(result_array.coords[cond]) for cond in pot_conds)
    while not it.finished:
        indep_idx = []
        # Relies on being ordered
        for sv in state_variables:
            if str(sv) in result_array.coords.keys():
                coord_idx = list(result_array.coords.keys()).index(str(sv))
                indep_idx.append(it.multi_index[coord_idx])
            else:
                # free state variable
                indep_idx.append(0)
        indep_idx = tuple(indep_idx)
        if len(comp_conds) > 0:
            comp_idx = np.ravel_multi_index(tuple(idx for idx, key in zip(it.multi_index, result_array_GM_dims) if key in comp_conds), comp_coord_shape)
            idx_comp_values = comp_values[comp_idx, :]
        else:
            idx_comp_values = np.atleast_1d(1.)
        if len(pot_conds) > 0:
            pot_idx = np.ravel_multi_index(tuple(idx for idx, key in zip(it.multi_index, result_array_GM_dims) if key in pot_conds), pot_coord_shape)
            idx_pot_values = np.array(cart_pot_values[pot_idx, :])

        idx_global_grid_X_values = global_grid_X_values[indep_idx]
        idx_global_grid_GM_values = global_grid_GM_values[indep_idx]
        idx_result_array_MU_values = result_array_MU_values[it.multi_index]
        idx_result_array_MU_values[:] = 0
        for idx in range(len(pot_conds_indices)):
            idx_result_array_MU_values[pot_conds_indices[idx]] = idx_pot_values[idx]
        idx_result_array_NP_values = result_array_NP_values[it.multi_index]
        idx_result_array_points_values = result_array_points_values[it.multi_index]
        result_array_GM_values[it.multi_index] = \
            hyperplane(idx_global_grid_X_values, idx_global_grid_GM_values,
                       idx_comp_values, idx_result_array_MU_values, float(global_grid.coords['N'][0]),
                       pot_conds_indices, comp_conds_indices,
                       idx_result_array_NP_values, idx_result_array_points_values)
        # Copy phase values out
        points = result_array_points_values[it.multi_index]
        result_array_Phase_values[it.multi_index][:num_comps] = global_grid_Phase_values[indep_idx].take(points, axis=0)[:num_comps]
        result_array_X_values[it.multi_index][:num_comps] = global_grid_X_values[indep_idx].take(points, axis=0)[:num_comps]
        result_array_Y_values[it.multi_index][:num_comps] = global_grid_Y_values[indep_idx].take(points, axis=0)[:num_comps]
        # Special case: Sometimes fictitious points slip into the result
        if '_FAKE_' in result_array_Phase_values[it.multi_index]:
            new_energy = 0.
            molesum = 0.
            for idx in range(len(result_array_Phase_values[it.multi_index])):
                midx = it.multi_index + (idx,)
                if result_array_Phase_values[midx] == '_FAKE_':
                    result_array_Phase_values[midx] = ''
                    result_array_X_values[midx] = np.nan
                    result_array_Y_values[midx] = np.nan
                    idx_result_array_NP_values[idx] = np.nan
                else:
                    new_energy += idx_result_array_NP_values[idx] * global_grid.GM[np.index_exp[indep_idx + (points[idx],)]]
                    molesum += idx_result_array_NP_values[idx]
            result_array_GM_values[it.multi_index] = new_energy / molesum
        it.iternext()
    result_array.remove('points')
    return result_array



import pycalphad.variables as v
from pycalphad.codegen.sympydiff_utils import build_functions
from pycalphad.core.utils import get_pure_elements, unpack_components, \
    extract_parameters, get_state_variables, wrap_symbol
from pycalphad.core.phase_rec import PhaseRecord
from pycalphad.core.constraints import build_constraints
from itertools import repeat
import warnings


def build_callables(dbf, comps, phases, models, parameter_symbols=None,
                    output='GM', build_gradients=True, build_hessians=False,
                    additional_statevars=None):
    """
    Create a compiled callables dictionary.

    Parameters
    ----------
    dbf : Database
        A Database object
    comps : list
        List of component names
    phases : list
        List of phase names
    models : dict
        Dictionary of {phase_name: Model subclass}
    parameter_symbols : list, optional
        List of string or SymEngine Symbols that will be overridden in the callables.
    output : str, optional
        Output property of the particular Model to sample. Defaults to 'GM'
    build_gradients : bool, optional
        Whether or not to build gradient functions. Defaults to True.
    build_hessians : bool, optional
        Whether or not to build Hessian functions. Defaults to False.
    additional_statevars : set, optional
        State variables to include in the callables that may not be in the models (e.g. from conditions)
    verbose : bool, optional
        Print the name of the phase when its callables are built

    Returns
    -------
    callables : dict
        Dictionary of keyword argument callables to pass to equilibrium.
        Maps {'output' -> {'function' -> {'phase_name' -> AutowrapFunction()}}.

    Notes
    -----
    *All* the state variables used in calculations must be specified.
    If these are not specified as state variables of the models (e.g. often the
    case for v.N), then it must be supplied by the additional_statevars keyword
    argument.

    Examples
    --------
    >>> from pycalphad import Database, equilibrium, variables as v
    >>> from pycalphad.codegen.callables import build_callables
    >>> from pycalphad.core.utils import instantiate_models
    >>> dbf = Database('AL-NI.tdb')
    >>> comps = ['AL', 'NI', 'VA']
    >>> phases = ['LIQUID', 'AL3NI5', 'AL3NI2', 'AL3NI']
    >>> models = instantiate_models(dbf, comps, phases)
    >>> callables = build_callables(dbf, comps, phases, models, additional_statevars={v.P, v.T, v.N})
    >>> 'GM' in callables.keys()
    True
    >>> 'massfuncs' in callables['GM']
    True
    >>> conditions = {v.P: 101325, v.T: 2500, v.X('AL'): 0.2}
    >>> equilibrium(dbf, comps, phases, conditions, callables=callables)
    """
    additional_statevars = set(additional_statevars) if additional_statevars is not None else set()
    parameter_symbols = parameter_symbols if parameter_symbols is not None else []
    parameter_symbols = sorted([wrap_symbol(x) for x in parameter_symbols], key=str)
    comps = sorted(unpack_components(dbf, comps))
    pure_elements = get_pure_elements(dbf, comps)

    _callables = {
        'massfuncs': {},
        'massgradfuncs': {},
        'masshessfuncs': {},
        'formulamolefuncs': {},
        'formulamolegradfuncs': {},
        'formulamolehessfuncs': {},
        'callables': {},
        'grad_callables': {},
        'hess_callables': {},
        'internal_cons_func': {},
        'internal_cons_jac': {},
        'internal_cons_hess': {},
    }

    state_variables = get_state_variables(models=models)
    state_variables |= additional_statevars
    if not {v.T, v.P, v.N}.issubset(state_variables):
        warnings.warn("State variables in `build_callables` are not {{N, P, T}}, but {}. This can lead to incorrectly "
                      "calculated values if the state variables used to call the generated functions do not match the "
                      "state variables used to create them. State variables can be added with the "
                      "`additional_statevars` argument.".format(state_variables))
    state_variables = sorted(state_variables, key=str)

    for name in phases:
        mod = models[name]
        site_fracs = mod.site_fractions
        try:
            out = getattr(mod, output)
        except AttributeError:
            raise AttributeError('Missing Model attribute {0} specified for {1}'
                                 .format(output, mod.__class__))

        # Build the callables of the output
        # Only force undefineds to zero if we're not overriding them
        undefs = {x for x in out.free_symbols if not isinstance(x, v.StateVariable)} - set(parameter_symbols)
        undef_vals = repeat(0., len(undefs))
        out = out.xreplace(dict(zip(undefs, undef_vals)))
        build_output = build_functions(out, tuple(state_variables + site_fracs), parameters=parameter_symbols,
                                       include_grad=build_gradients, include_hess=build_hessians)
        cf, gf, hf = build_output.func, build_output.grad, build_output.hess
        _callables['callables'][name] = cf
        _callables['grad_callables'][name] = gf
        _callables['hess_callables'][name] = hf

        # Build the callables for mass
        # TODO: In principle, we should also check for undefs in mod.moles()
        mcf, mgf, mhf = zip(*[build_functions(mod.moles(el), state_variables + site_fracs,
                                              include_obj=True,
                                              include_grad=build_gradients,
                                              include_hess=build_hessians,
                                              parameters=parameter_symbols)
                              for el in pure_elements])

        _callables['massfuncs'][name] = mcf
        _callables['massgradfuncs'][name] = mgf
        _callables['masshessfuncs'][name] = mhf

        # Build the callables for moles per formula unit
        # TODO: In principle, we should also check for undefs in mod.moles()
        fmcf, fmgf, fmhf = zip(*[build_functions(mod.moles(el, per_formula_unit=True), state_variables + site_fracs,
                                                 include_obj=True,
                                                 include_grad=build_gradients,
                                                 include_hess=build_hessians,
                                                 parameters=parameter_symbols)
                                 for el in pure_elements])

        _callables['formulamolefuncs'][name] = fmcf
        _callables['formulamolegradfuncs'][name] = fmgf
        _callables['formulamolehessfuncs'][name] = fmhf
    return {output: _callables}


def build_phase_records(dbf, comps, phases, state_variables, models, output='GM',
                        callables=None, parameters=None, verbose=False,
                        build_gradients=True, build_hessians=True
                        ):
    """
    Combine compiled callables and callables from conditions into PhaseRecords.

    Parameters
    ----------
    dbf : Database
        A Database object
    comps : List[Union[str, v.Species]]
        List of active pure elements or species.
    phases : list
        List of phase names
    state_variables : Iterable[v.StateVariable]
        State variables used to produce the generated functions.
    models : Mapping[str, Model]
        Mapping of phase names to model instances
    parameters : dict, optional
        Maps SymEngine Symbol to numbers, for overriding the values of parameters in the Database.
    callables : dict, optional
        Pre-computed callables. If None are passed, they will be built.
        Maps {'output' -> {'function' -> {'phase_name' -> AutowrapFunction()}}
    output : str
        Output property of the particular Model to sample
    verbose : bool, optional
        Print the name of the phase when its callables are built
    build_gradients : bool
        Whether or not to build gradient functions. Defaults to False. Only
        takes effect if callables are not passed.
    build_hessians : bool
        Whether or not to build Hessian functions. Defaults to False. Only
        takes effect if callables are not passed.

    Returns
    -------
    dict
        Dictionary mapping phase names to PhaseRecord instances.

    Notes
    -----
    If callables are passed, don't rebuild them. This means that the callables
    are not checked for incompatibility. Users of build_callables are
    responsible for ensuring that the state variables, parameters and models
    used to construct the callables are compatible with the ones used to
    build the constraints and phase records.

    """
    comps = sorted(unpack_components(dbf, comps))
    parameters = parameters if parameters is not None else {}
    callables = callables if callables is not None else {}
    _constraints = {
        'internal_cons_func': {},
        'internal_cons_jac': {},
        'internal_cons_hess': {},
    }
    phase_records = {}
    state_variables = sorted(get_state_variables(models=models, conds=state_variables), key=str)
    param_symbols, param_values = extract_parameters(parameters)

    if callables.get(output) is None:
        callables = build_callables(dbf, comps, phases, models,
                                    parameter_symbols=parameters.keys(), output=output,
                                    additional_statevars=state_variables,
                                    build_gradients=False,
                                    build_hessians=False)
    # Temporary solution. PhaseRecord needs rework: https://github.com/pycalphad/pycalphad/pull/329#discussion_r634579356
    formulacallables = build_callables(dbf, comps, phases, models,
                                       parameter_symbols=parameters.keys(), output='G',
                                       additional_statevars=state_variables,
                                       build_gradients=build_gradients,
                                       build_hessians=build_hessians)

    # If a vector of parameters is specified, only pass the first row to the PhaseRecord
    # Future callers of PhaseRecord.obj_parameters_2d() can pass the full param_values array as an argument
    if len(param_values.shape) > 1:
        param_values = param_values[0]

    for name in phases:
        mod = models[name]
        site_fracs = mod.site_fractions
        # build constraint functions
        cfuncs = build_constraints(mod, state_variables + site_fracs, parameters=param_symbols)
        _constraints['internal_cons_func'][name] = cfuncs.internal_cons_func
        _constraints['internal_cons_jac'][name] = cfuncs.internal_cons_jac
        _constraints['internal_cons_hess'][name] = cfuncs.internal_cons_hess
        num_internal_cons = cfuncs.num_internal_cons

        phase_records[name.upper()] = PhaseRecord(comps, state_variables, site_fracs, param_values,
                                                  callables[output]['callables'][name],
                                                  formulacallables['G']['callables'][name],
                                                  formulacallables['G']['grad_callables'][name],
                                                  formulacallables['G']['hess_callables'][name],
                                                  callables[output]['massfuncs'][name],
                                                  formulacallables['G']['formulamolefuncs'][name],
                                                  formulacallables['G']['formulamolegradfuncs'][name],
                                                  formulacallables['G']['formulamolehessfuncs'][name],
                                                  _constraints['internal_cons_func'][name],
                                                  _constraints['internal_cons_jac'][name],
                                                  _constraints['internal_cons_hess'][name],
                                                  num_internal_cons)

        if verbose:
            print(name + ' ')
    return phase_records



# distutils: language = c++
from pycalphad.core.phase_rec cimport PhaseRecord

cdef public class CompositionSet(object)[type CompositionSetType, object CompositionSetObject]:
    cdef public PhaseRecord phase_record
    cdef readonly double[::1] dof, X
    cdef double[:,::1] _X_2d_view
    cdef public double NP
    cdef public bint fixed
    cdef readonly double energy
    cdef double[::1] _energy_2d_view
    cpdef void update(self, double[::1] site_fracs, double phase_amt, double[::1] state_variables)



import itertools
from typing import List, Union
from collections import OrderedDict
from functools import partial
from symengine import log, S, Symbol
from tinydb import where
from pycalphad.model import _MAX_PARAM_NESTING
import pycalphad.variables as v
from pycalphad.core.utils import unpack_components, wrap_symbol
from pycalphad import Model
from pycalphad.core.errors import DofError


def _get_species(i, j, k, l) -> v.Species:
    """Return a Species for quadruplet given by constituents

    Canonicalizes the Species by sorting among the cation and anion sublattices.

    Parameters
    ----------
    constituents : List[Union[str, v.Species]]

    """
    constituents = [i, j, k, l]
    if all(isinstance(c, v.Species) for c in constituents):
        # if everything is a species, get the string names as constituents
        constituents = [c.name for c in constituents]
    if len(constituents) == 4:
        constituents = sorted(constituents[0:2]) + sorted(constituents[2:4])
    name = "".join(constituents)
    constituent_dict = {}
    # using get() will increment c instead of overriding if c is already defined
    for c in constituents:
        constituent_dict[c] = constituent_dict.get(c, 0.0) + 1.0
    return v.Species(name, constituents=constituent_dict)


class ModelMQMQA(Model):
    """
    Symbolic implementation of the modified quasichemical model in the
    quadruplet approximation developed by Pelton _et al._ [1]_. The formulation
    here largely follows the derivation by Poschmann _et al._ [2]_.

    Following pycalphad convention, ``ModelMQMQA.G`` defines the energy for one
    "formula unit" of this phase. The MQMQA formulations by Pelton _et al._ and
    Poschmann _et al._ do not formally define the energy per formula unit in
    the same way as is done for the CEF. However, it is convienient to define
    one formula unit of an MQMQA phase to be the energy corresponding to one
    mole of quadruplets. In terms of the equations from Poschmann _et al._,
    which are largely followed in this class, that means
    :math:`\\sum_{ab/xy} n_{ab/xy} = 1` and therefore
    :math:`n_{ab/xy} = X_{ij/kl}` by definition.

    This class is only semantically a subclass of ``Model``. It implements the
    API expected for a Model in a self-contained way without any need to rely
    on the Model superclass, but it is created as a subclass to satisfy various
    ``isinstance`` checks in the codebase. The subclassing on Model should be
    removed once a suitable abstract base class or protocol is defined.

    References
    ----------
    .. [1] Pelton, Chartrand, and Eriksson, The Modified Quasi-chemical Model: Part IV. Two-Sublattice Quadruplet Approximation, Metallurgical and Materials Transactions A, 32(6) (2001) 1409-1416 doi: `10.1007/s11661-001-0230-7 <https://doi.org/10.1007/s11661-001-0230-7>`_
    .. [2] Poschmann, Bajpai, Fitzpatrick, and Piro, Recent developments for molten salt systems in Thermochimica, CALPHAD 75 (2021) 102341 doi: `10.1016/j.calphad.2021.102341 <https://doi.org/10.1016/j.calphad.2021.102341>`_

    """

    contributions = [
        ("ref", "reference_energy"),
        ("idmix", "ideal_mixing_energy"),
        ("xsmix", "excess_mixing_energy"),
    ]

    def __init__(self, dbe, comps, phase_name, parameters=None):
        self._dbe = dbe
        self._reference_model = None
        self.components = set()
        self.constituents = []
        self.phase_name = phase_name.upper()
        phase = dbe.phases[self.phase_name]
        self.site_ratios = tuple(list(phase.sublattices))

        active_species = unpack_components(dbe, comps)
        constituents = []
        for sublattice in dbe.phases[phase_name].constituents:
            sublattice_comps = set(sublattice).intersection(active_species)
            self.components |= sublattice_comps
            constituents.append(sublattice_comps)
        self.components = sorted(self.components)
        # create self.cations and self.anions properties to use instead of constituents
        if len(constituents) == 2:
            self.cations = sorted(constituents[0])
            self.anions = sorted(constituents[1])
        else:
            raise ValueError(f"Exactly two sublattices are required for the MQMQA. Got {len(constituents)} sublattices: {constituents}.")

        # In several places we use the assumption that the cation lattice and
        # anion lattice have no common species; we validate that assumption here
        shared_species = set(self.cations).intersection(set(self.anions))
        assert len(shared_species) == 0, f"No species can be shared between the two MQMQA lattices, got {shared_species}"

        # Verify that this phase is still possible to build
        if len(self.cations) == 0:
            raise DofError(f"{self.phase_name}: Cation sublattice of {phase.constituents[0]} has no active species in {self.components}")
        if len(self.anions) == 0:
            raise DofError(f"{self.phase_name}: Anion sublattice of {phase.constituents[1]} has no active species in {self.components}")

        self._pairs = list(itertools.product(self.cations, self.anions))
        quads = list(itertools.product(
            itertools.combinations_with_replacement(self.cations, 2),
            itertools.combinations_with_replacement(self.anions, 2),
        ))
        self._quadruplets = [(A, B, X, Y) for (A, B), (X, Y) in quads]
        quad_species = [_get_species(A, B, X, Y) for (A, B), (X, Y) in quads]
        self.constituents = [sorted(quad_species)]

        # Convert string symbol names to symengine Symbol objects
        # This makes xreplace work with the symbols dict
        symbols = {Symbol(s): val for s, val in dbe.symbols.items()}

        if parameters is not None:
            self._parameters_arg = parameters
            if isinstance(parameters, dict):
                symbols.update([(wrap_symbol(s), val) for s, val in parameters.items()])
            else:
                # Lists of symbols that should remain symbolic
                for s in parameters:
                    symbols.pop(wrap_symbol(s))
        else:
            self._parameters_arg = None

        self._symbols = {wrap_symbol(key): value for key, value in symbols.items()}

        self.models = OrderedDict()
        self.build_phase(dbe)

        for name, value in self.models.items():
            self.models[name] = self.symbol_replace(value, symbols)

    def __eq__(self, other):
        if self is other:
            return True
        elif type(self) != type(other):
            return False
        else:
            return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(repr(self))

    # Default methods, don't need to be overridden unless you want to change the behavior
    @property
    def state_variables(self) -> List[v.StateVariable]:
        """Return a sorted list of state variables used in the ast which are not site fractions."""
        return sorted((x for x in self.ast.free_symbols if not isinstance(x, v.SiteFraction) and isinstance(x, v.StateVariable)), key=str)

    @property
    def site_fractions(self) -> List[v.SiteFraction]:
        """Return a sorted list of site fractions used in the ast."""
        return sorted((x for x in self.ast.free_symbols if isinstance(x, v.SiteFraction)), key=str)

    @property
    def ast(self):
        "Return the full abstract syntax tree of the model."
        return sum(self.models.values())

    def _pair_test(self, constituent_array):
        """Return True if the constituent array represents a pair.

        Pairs have only one species in each sublattice.
        """
        for subl in constituent_array:
            if len(subl) > 1:
                return False
            constituent = subl[0]
            if constituent not in self.components:
                return False
        return True

    def _array_validity(self, constituent_array):
        """Return True if the constituent array is satisfies all components."""
        for subl in constituent_array:
            for constituent in subl:
                if constituent not in self.components:
                    return False
        return True

    def _zeta_ik(self, i, k):
        pair_query = (
            (where("phase_name") == self.phase_name) & \
            (where("parameter_type") == "MQMG") & \
            (where("constituent_array").test(lambda x: x == ((i,), (k,))))
        )
        params = list(self._dbe._parameters.search(pair_query))
        assert len(params) == 1, f"Expected exactly one pair parameter for ({i, k}), got {len(params)}: {params}"
        param = params[0]
        return param["zeta"]

    def _X_ijkl(self, i, j, k, l) -> v.SiteFraction:
        """
        Shorthand for creating a site fraction object v.Y for a quadruplet (ij/kl)
        """
        return v.Y(self.phase_name, 0, _get_species(i, j, k, l))

    def _n_ik(self, i, k):
        """
        Return the amount of pair, n_i/k, for the pair i/k.

        Follows Poschmann Eq. 5 except that including zeta here (as by
        Poschmann) seems to break the tests when the coordination numbers
        vary across quadruplets.
        """
        n_ik = S.Zero
        for a, b, x, y in self._quadruplets:
            n_ik += self._X_ijkl(a, b, x, y) * ((a == i) + (b == i)) * ((x == k) + (y == k))
        return n_ik

    def _n_ik_star(self, i, k):
        """
        Return the amount of pair, n^*_i/k, for the pair i/k following Lambotte JCT 2011 Eq. A.3

        This term is only used in the pair part of the entropy.
        """
        n_ik = S.Zero
        for a, b, x, y in self._quadruplets:
            n_ik += self._X_ijkl(a, b, x, y) * ((a == i) + (b == i)) * ((x == k) + (y == k))
        n_ik /= self._zeta_ik(i, k)
        return n_ik

    def _X_ik(self, i: v.Species, k: v.Species):
        """
        Return the endmember fraction, X_i/k, for a the pair i/k following Poschmann Eq. 6
        """
        X_ik = self._n_ik(i, k) / sum(self._n_ik(a, x) for a, x in self._pairs)
        return X_ik

    def _X_ik_star(self, i: v.Species, k: v.Species):
        """
        Return the endmember fraction, X^*_i/k, for a the pair i/k.

        Follows Lambotte JCT 2011 Eq. A.5, but this is simply Poschmann Eq. 6
        with n^*_ik instead of n_ik.

        This term is only used in the pair part of the entropy and as a
        sub-expression of F_i (also part of the pair part of the entropy).
        """
        X_ik = self._n_ik_star(i, k) / sum(self._n_ik_star(a, x) for a, x in self._pairs)
        return X_ik

    def _n_i(self, dbe, species):
        """
        Return the mass of the species following Poschmann Eq. 7 and 8.
        """
        Z = partial(self.Z, dbe)  # alias for notation
        n_i = S.Zero
        if species in self.cations:
            i = species
            for a, b, x, y in self._quadruplets:
                n_i += self._X_ijkl(a, b, x, y) * ((a == i) / Z(a, a, b, x, y) + (b == i) / Z(b, a, b, x, y))
        else:
            assert species in self.anions
            k = species
            for a, b, x, y in self._quadruplets:
                n_i += self._X_ijkl(a, b, x, y) * ((x == k) / Z(x, a, b, x, y) + (y == k) / Z(y, a, b, x, y))
        return n_i

    def _X_i(self, dbe, species: v.Species):
        """
        Return the site fraction of species on it's sublattice. Poschmann Eq. 9 and 10.
        """
        if species in self.cations:
            return self._n_i(dbe, species) / sum(self._n_i(dbe, a) for a in self.cations)
        else:
            assert species in self.anions
            return self._n_i(dbe, species) / sum(self._n_i(dbe, x) for x in self.anions)

    def _Y_i(self, species: v.Species):
        """
        Return the site equivalent fraction of species following Poschmann Eq. 11 and 12.
        """
        Y_i = S.Zero
        if species in self.cations:
            i = species
            for a, b, x, y in self._quadruplets:
                Y_i += self._X_ijkl(a, b, x, y) * ((a == i) + (b == i)) / 2
        else:
            assert species in self.anions
            k = species
            for a, b, x, y in self._quadruplets:
                Y_i += self._X_ijkl(a, b, x, y) * ((x == k) + (y == k)) / 2
        return Y_i

    def _F_i(self, species: v.Species):
        """
        Return the coordination-equivalent fraction of species, Poschmann Eq. 13 and 14.
        """
        soln_type = self._dbe.phases[self.phase_name].model_hints["mqmqa"]["type"]
        F_i = S.Zero
        if species in self.cations:
            i = species
            for a, x in self._pairs:
                if soln_type == "SUBG":
                    F_i += int(a == i) * self._X_ik(a, x)
                elif soln_type == "SUBQ":
                    F_i += int(a == i) * self._X_ik_star(a, x)
        else:
            assert species in self.anions
            k = species
            for a, x in self._pairs:
                if soln_type == "SUBG":
                    F_i += int(x == k) * self._X_ik(a, x)
                elif soln_type == "SUBQ":
                    F_i += int(x == k) * self._X_ik_star(a, x)
        return F_i


    def _chemical_group_filter(self, dbe, symmetric_species, asymmetric_species, sublattice):
        """
        Return a function ``f(m)`` that returns ``True`` if m is symmetric with
        the symmetric_species and asymmetric with the asymmetric_species.
        """
        # sublattice should be "cations" or "anions"
        chem_group_dict = dbe.phases[self.phase_name].model_hints["mqmqa"]["chemical_groups"][sublattice]
        def _f(species):
            if species == symmetric_species:
                return False
            elif species == asymmetric_species:
                return False
            elif chem_group_dict[species] == chem_group_dict[symmetric_species] and chem_group_dict[species] != chem_group_dict[asymmetric_species]:
                return True  # This chemical group should be mixed
            else:
                return False
        return _f

    def _Chi_mix(self, dbe, i, j, k, l):
        """
        (:math:`\\chi_{ij/k}`) following Poschmann Eq. 21 (SUBG-type model) or Eq. 22 (SUBQ-type models), respectively.
        """
        cations = self.cations
        anions = self.anions
        X_ijkl = self._X_ijkl
        soln_type = dbe.phases[self.phase_name].model_hints["mqmqa"]["type"]

        mixing_term_numerator = S.Zero
        mixing_term_denominator = S.Zero

        if i == j and k == l:
            raise ValueError(f"Excess energies for pairs are not defined. Got quadruplet {(i, j, k, l)}")
        elif i != j and k == l:  # Mixing on first sublattice
            # Chi_{ij/kk} type
            nu = list(filter(self._chemical_group_filter(dbe, i, j, "cations"), cations))
            gamma = list(filter(self._chemical_group_filter(dbe, j, i, "cations"), cations))
            for idx, a in enumerate([i] + nu):  # enumerate to avoid double counting
                for b in ([i] + nu)[idx:]:
                    if soln_type == "SUBG":
                        mixing_term_numerator += X_ijkl(a, b, k, k)
                    elif soln_type == "SUBQ":
                        # Eq. 22 loop over anions
                        for x in anions:
                            for y in anions:
                                mixing_term_numerator += X_ijkl(a, b, x, y) * ((k == x) + (k == y)) / 2
                    else:
                        raise ValueError(f"Unknown solution type: {soln_type}")
            for idx, a in enumerate([i, j] + nu + gamma):  # enumerate to avoid double counting
                for b in ([i, j] + nu + gamma)[idx:]:
                    if soln_type == "SUBG":
                        mixing_term_denominator += X_ijkl(a, b, k, k)
                    elif soln_type == "SUBQ":
                        # Eq. 22 loop over anions
                        for x in anions:
                            for y in anions:
                                mixing_term_denominator += X_ijkl(a, b, x, y) * ((k == x) + (k == y)) / 2
                    else:
                        raise ValueError(f"Unknown solution type: {soln_type}")
            return mixing_term_numerator / mixing_term_denominator
        elif i == j and k != l: # Mixing on second sublattice
            # Chi_{ii/kl} type
            nu = list(filter(self._chemical_group_filter(dbe, k, l, "anions"), anions))
            gamma = list(filter(self._chemical_group_filter(dbe, l, k, "anions"), anions))
            for idx, x in enumerate([k] + nu):  # enumerate to avoid double counting
                for y in ([k] + nu)[idx:]:
                    if soln_type == "SUBG":
                        mixing_term_numerator += X_ijkl(i, i, x, y)
                    elif soln_type == "SUBQ":
                        # Eq. 22 loop over cations
                        for a in cations:
                            for b in cations:
                                mixing_term_numerator += X_ijkl(a, b, x, y) * ((i == a) + (i == b)) / 2
                    else:
                        raise ValueError(f"Unknown solution type: {soln_type}")
            for idx, x in enumerate([k, l] + nu + gamma):  # enumerate to avoid double counting
                for y in ([k, l] + nu + gamma)[idx:]:
                    if soln_type == "SUBG":
                        mixing_term_denominator += X_ijkl(i, i, x, y)
                    elif soln_type == "SUBQ":
                        # Eq. 22 loop over cations
                        for a in cations:
                            for b in cations:
                                mixing_term_denominator += X_ijkl(a, b, x, y) * ((i == a) + (i == b)) / 2
                    else:
                        raise ValueError(f"Unknown solution type: {soln_type}")
            return mixing_term_numerator / mixing_term_denominator
        else:
            raise ValueError(f"Computing Chi_mix is not supported for reciprocal quadruplets. Got quadruplet {(i, j, k, l)}.")

    def _Y_ik(self, i, k):
        """
        Poschmann Eq. 20
        """
        term = S.Zero
        for a, b, x, y in self._quadruplets:
            term += self._X_ijkl(a,b,x,y) * ((a == i) + (b == i)) * ((x == k) + (y == k)) / 4
        return term

    def _Xi_mix(self, dbe, i, j, k, l):
        """
        (:math:`\\xi_{ij/k}`) following Poschmann Eq. 19
        """
        # For mixing in cations (i != j, k == l), nu are cations (nu != i, nu != j) where i and nu have the same
        # chemical group and j has a different chemical group.
        mixing_term = S.Zero
        if i == j and k == l:
            raise ValueError(f"Computing Xi_mix is not supported for pair quadruplets (there must be mixing among cations or anions). Got quadruplet {(i, j, k, l)}.")
        elif i != j and k == l:  # Mixing on first sublattice
            nu = list(filter(self._chemical_group_filter(dbe, i, j, "cations"), self.cations))
            for a in [i] + nu:
                mixing_term += self._Y_ik(a, k)
            return mixing_term
        elif i == j and k != l:  # Mixing on second sublattice
            nu = list(filter(self._chemical_group_filter(dbe, k, l, "anions"), self.anions))
            for x in [k] + nu:
                mixing_term += self._Y_ik(i, x)
            return mixing_term
        else:
            raise ValueError(f"Computing Xi_mix is not supported for reciprocal quadruplets (there can only be mixing among cations _or_ anions). Got quadruplet {(i, j, k, l)}.")

    def _calc_Z(self, dbe, species, A, B, X, Y):
        # In derivations of the MQMQA, charges are written as if they have the same sign. The absolute values of the charges are used here.
        Z = partial(self.Z, dbe)
        if (species == A) or (species == B):
            species_is_cation = True
        elif (species == X) or (species == Y):
            species_is_cation = False
        else:
            raise ValueError(f"{species} is not A ({A}), B ({B}), X ({X}) or Y ({Y}).")

        if A == B and X == Y:
            raise ValueError(f"Z({species}, {A}{B}/{X}{Y}) is a pure pair and must be defined explictly")
        elif A != B and X != Y:
            # This is a reciprocal AB/XY quadruplet and needs to be calculated by eq 23 and 24 in Pelton et al. Met Trans B (2001)
            F = 1/8 * (  # eq. 24
                  abs(A.charge) / Z(A, A,A,X,Y)
                + abs(B.charge) / Z(B, B,B,X,Y)
                + abs(X.charge) / Z(X, A,B,X,X)
                + abs(Y.charge) / Z(Y, A,B,Y,Y)
                )
            if species_is_cation:
                inv_Z = F * (
                    Z(X, A,B,X,X) / (abs(X.charge) * Z(species, A,B,X,X))
                    + Z(Y, A,B,Y,Y) / (abs(Y.charge) * Z(species, A,B,Y,Y))
                            )

            else:
                inv_Z = F * (
                    Z(A, A,A,X,Y) / (abs(A.charge) * Z(species, A,A,X,Y))
                    + Z(B, B,B,X,Y) / (abs(B.charge) * Z(species, B,B,X,Y))
                            )

            return 1 / inv_Z
        elif A != B:  # X == Y
            # Need to calculate Z^i_AB/XX (Y = X).
            # We assume Z^A_ABXX = Z^A_AAXX = Z^A_AAYY
            # and Z^X_ABXX = (q_X + q_Y)/(q_A/Z^A_AAXX + q_B/Z^B_BBXX)  # note: q_X = q_Y, etc. since Y = X
            # We don't know if these are correct, but that's what's implemented in Thermochimica
            if species_is_cation:
                return Z(species, species, species, X, X)
            else:
                return 2*abs(species.charge) / (
                    abs(A.charge) / Z(A, A,A,species,species) + abs(B.charge) / Z(B, B,B,species,species)
                    )
        elif X != Y:  # A == B
            # These use the same equations as A != B case with the same assumptions
            if species_is_cation:
                # similarly, Z^A_AAXY = (q_A + q_B)/(q_X/Z^X_AAXX + q_Y/Z^Y_AAYY)
                return 2*abs(species.charge)/(abs(X.charge)/Z(X, species, species, X, X) + abs(Y.charge)/Z(Y, species, species, Y, Y))
            else:
                return Z(species, A, A, species, species)
        raise ValueError("This should be unreachable")


    def Z(self, dbe, species: v.Species, A: v.Species, B: v.Species, X: v.Species, Y: v.Species):
        # Canonicalize the order of cations and anions in alphabetical order
        # Note that the coordination number parameters (the constituents and
        # the values) _must_ obey this canonical sorted order.
        A, B = sorted((A, B))
        X, Y = sorted((X, Y))
        Zs = dbe._parameters.search(
            (where("phase_name") == self.phase_name) & \
            (where("parameter_type") == "MQMZ") & \
            (where("constituent_array").test(lambda x: x == ((A, B), (X, Y))))
        )
        if len(Zs) == 0:
            return self._calc_Z(dbe, species, A, B, X, Y)
        elif len(Zs) == 1:
            sp_idx = [A, B, X, Y].index(species)
            return Zs[0]["coordinations"][sp_idx]
        else:
            raise ValueError(f"Expected exactly one Z for {species} of {((A, B), (X, Y))}, got {len(Zs)}")

    def get_internal_constraints(self):
        constraints = []
        # Site fraction constraint
        site_frac_sum = S.Zero
        for a, b, x, y in self._quadruplets:
            site_frac_sum += self._X_ijkl(a, b, x, y)
        constraints.append(site_frac_sum - 1)  # = 0
        return constraints

    @property
    def normalization(self):
        """
        Total number of moles of this phase.

        Divide by this normalization factor to convert from J/mole-formula to J/mole-atoms
        """
        return sum(self._n_i(self._dbe, i) for i in self.cations + self.anions if "VA" not in i.constituents)

    def moles(self, species, per_formula_unit=False):
        "Number of moles of species or elements."
        species = v.Species(species)
        result = S.Zero
        for i in self.cations + self.anions:
            if list(species.constituents.keys())[0] in i.constituents:
                result += self._n_i(self._dbe, i)
        if per_formula_unit:
            return result
        else:
            return result / self.normalization

    degree_of_ordering = DOO = S.Zero
    curie_temperature = TC = S.Zero
    beta = BMAG = S.Zero
    neel_temperature = NT = S.Zero

    # pylint: disable=C0103
    # These are standard abbreviations from Thermo-Calc for these quantities
    GM = property(lambda self: self.ast / self.normalization)
    G = property(lambda self: self.ast)
    energy = GM
    entropy = SM = property(lambda self: -self.GM.diff(v.T))
    enthalpy = HM = property(lambda self: self.GM - v.T * self.GM.diff(v.T))
    heat_capacity = CPM = property(lambda self: -v.T * self.GM.diff(v.T, v.T))
    # pylint: enable=C0103
    mixing_energy = GM_MIX = property(lambda self: self.GM - self.reference_model.GM)
    mixing_enthalpy = HM_MIX = property(lambda self: self.GM_MIX - v.T * self.GM_MIX.diff(v.T))
    mixing_entropy = SM_MIX = property(lambda self: -self.GM_MIX.diff(v.T))
    mixing_heat_capacity = CPM_MIX = property(lambda self: -v.T * self.GM_MIX.diff(v.T, v.T))

    def reference_energy(self, dbe):
        """
        Returns energies contributed by pairs.
        """
        pair_query = (
            (where("phase_name") == self.phase_name) & \
            (where("parameter_type") == "MQMG") & \
            (where("constituent_array").test(self._pair_test))
        )
        params = dbe._parameters.search(pair_query)
        terms = S.Zero
        for param in params:
            a = i = param["constituent_array"][0][0]
            x = k = param["constituent_array"][1][0]
            X_ax = S.Zero
            for b in self.cations:
                for y in self.anions:
                    X_ax += self._X_ijkl(a,b,x,y) * ((a == i) + (b == i)) * ((x == k) + (y == k)) / (2 * self.Z(dbe, a, a,b,x,y))
            G_ax = param["parameter"]
            terms += X_ax * G_ax / param["stoichiometry"][0]
        return terms

    def ideal_mixing_energy(self, dbe):
        # notational niceties
        n_i = partial(self._n_i, dbe)
        X_i = partial(self._X_i, dbe)
        X_ik = self._X_ik
        Y_i = self._Y_i
        F_i = self._F_i
        X_ijkl = self._X_ijkl
        soln_type = dbe.phases[self.phase_name].model_hints["mqmqa"]["type"]
        if soln_type == "SUBG":
            pow_X_ik = 1.0
            pow_Y_i = 1.0
        elif soln_type == "SUBQ":
            pow_X_ik = 0.75
            pow_Y_i = 0.5

        Sid = S.Zero
        # Individual constituents
        for A in self.cations:
            Sid += n_i(A) * log(X_i(A))
        for X in self.anions:
            Sid += n_i(X) * log(X_i(X))
        # Pairs
        if soln_type == "SUBG":
            for i, k in self._pairs:
                Sid += self._n_ik(i, k) * log(X_ik(i, k) / (F_i(i) * F_i(k)))
        elif soln_type == "SUBQ":
            for i, k in self._pairs:
                Sid += self._n_ik_star(i, k) * log(self._X_ik_star(i, k) / (F_i(i) * F_i(k)))
        # Quadruplets
        for i, j, k, l in self._quadruplets:
            C_ijkl = (2 - (i == j)) * (2 - (k == l))
            Sid += X_ijkl(i, j, k, l) * log(X_ijkl(i, j, k, l) / (C_ijkl * (X_ik(i, k) * X_ik(i, l) * X_ik(j, k) * X_ik(j, l))**pow_X_ik / (Y_i(i) * Y_i(j) * Y_i(k) * Y_i(l))**pow_Y_i))
        return Sid * v.T * v.R

    def excess_mixing_energy(self, dbe):
        params = dbe._parameters.search(
            (where("phase_name") == self.phase_name) &
            (where("parameter_type") == "MQMX") &
            (where("constituent_array").test(self._array_validity))
        )

        cations = self.cations
        anions = self.anions

        X_ijkl = self._X_ijkl
        Z = partial(self.Z, dbe)

        energy = S.Zero
        for param in params:
            (A, B), (X, Y) = param["constituent_array"]
            exponents = param["exponents"]
            mixing_code = param["mixing_code"]
            m = param["additional_mixing_constituent"]
            p_alpha = exponents[0]
            q_alpha = exponents[1]
            r_alpha = param["additional_mixing_exponent"]
            # Poschmann Eq. 23-26
            mixing_term = S.Zero
            if A != B and X == Y:
                if mixing_code == "G":
                    # Poschmann Eq. 23 (cations mixing)
                    mixing_term += self._Chi_mix(dbe, A, B, X, X)**p_alpha * self._Chi_mix(dbe, B, A, X, X)**q_alpha
                elif mixing_code == "Q":
                    # Poschmann Eq. 19 and 20 (cations mixing)
                    Xi_ijk = self._Xi_mix(dbe, A, B, X, X)
                    Xi_jik = self._Xi_mix(dbe, B, A, X, X)
                    # Poschmann Eq. 24
                    mixing_term += Xi_ijk**p_alpha * Xi_jik**q_alpha / (Xi_ijk + Xi_jik)**(p_alpha + q_alpha)
                else:
                    raise ValueError(f"Unknown mixing code {mixing_code} for parameter {param}")
                if m != v.Species(None):
                    # Poschmann Eq. 25 and 26 ternary term (same for both mixing codes)
                    Xi_ijk = self._Xi_mix(dbe, A, B, X, X)
                    Xi_jik = self._Xi_mix(dbe, B, A, X, X)
                    Y_mk = self._Y_ik(m, X)
                    nu = list(filter(self._chemical_group_filter(dbe, A, B, "cations"), cations))
                    gamma = list(filter(self._chemical_group_filter(dbe, B, A, "cations"), cations))
                    if m in gamma:
                        mixing_term *= Y_mk / Xi_jik * (1 - self._Y_ik(B, X) / Xi_jik)**(r_alpha - 1)
                    elif m in nu:
                        mixing_term *= Y_mk / Xi_ijk * (1 - self._Y_ik(A, X) / Xi_ijk)**(r_alpha - 1)
                    else:  # not in nu or gamma
                        mixing_term *= Y_mk * (1 - Xi_ijk - Xi_jik)**(r_alpha - 1)
            elif A == B and X != Y:
                if mixing_code == "G":
                    # Poschmann Eq. 23 (anions mixing)
                    mixing_term += self._Chi_mix(dbe, A, A, X, Y)**p_alpha * self._Chi_mix(dbe, A, A, Y, X)**q_alpha
                elif mixing_code == "Q":
                    # Poschmann Eq. 19 and 20 (anions mixing)
                    Xi_ikl = self._Xi_mix(dbe, A, A, X, Y)
                    Xi_ilk = self._Xi_mix(dbe, A, A, Y, X)
                    # Poschmann Eq. 24
                    mixing_term += Xi_ikl**p_alpha * Xi_ilk**q_alpha / (Xi_ikl + Xi_ilk)**(p_alpha + q_alpha)
                else:
                    raise ValueError(f"Unknown mixing code {mixing_code} for parameter {param}")
                if m != v.Species(None):
                    # Poschmann Eq. 25 and 26 ternary term (same for both mixing codes)
                    Xi_ikl = self._Xi_mix(dbe, A, A, X, Y)
                    Xi_ilk = self._Xi_mix(dbe, A, A, Y, X)
                    Y_im = self._Y_ik(A, m)
                    nu = list(filter(self._chemical_group_filter(dbe, X, Y, "anions"), anions))
                    gamma = list(filter(self._chemical_group_filter(dbe, Y, X, "anions"), anions))
                    if m in gamma:
                        mixing_term *= Y_im / Xi_ilk * (1 - self._Y_ik(A, Y) / Xi_ilk)**(r_alpha - 1)
                    elif m in nu:
                        mixing_term *= Y_im / Xi_ikl * (1 - self._Y_ik(A, X) / Xi_ikl)**(r_alpha - 1)
                    else:  # not in nu or gamma
                        mixing_term *= Y_im * (1 - Xi_ikl - Xi_ilk)**(r_alpha - 1)
            else:
                mixing_term = S.One  # No mixing, this is a modification to the formation energy of this quadruplet
            g = param["parameter"] * mixing_term

            # Poschmann Eq. 17
            cation_factor = S.Zero
            if A == B:
                for m in cations:
                    if m != A:
                        cation_factor += X_ijkl(A,m,X,Y) / Z(A, A,m,X,Y)
                cation_factor *= Z(A, A,B,X,Y) / 2
            anion_factor = S.Zero
            if X == Y:
                for m in anions:
                    if m != X:
                        anion_factor += X_ijkl(A,B,X,m) / Z(X, A,B,X,m)
                anion_factor *= Z(X, A,B,X,Y) / 2
            energy += 0.5 * g * (X_ijkl(A,B,X,Y) + cation_factor + anion_factor)

        return energy

    def shift_reference_state(self, reference_states, dbe, contrib_mods=None, output=("GM", "HM", "SM", "CPM"), fmt_str="{}R"):
        raise NotImplementedError()

    def build_phase(self, dbe):
        """
        Generate the symbolic form of all the contributions to this phase.

        Parameters
        ----------
        dbe : 'pycalphad.io.Database'
        """
        self.models.clear()
        for key, value in self.__class__.contributions:
            self.models[key] = S(getattr(self, value)(dbe))

    @staticmethod
    def symbol_replace(obj, symbols):
        """
        Substitute values of symbols into 'obj'.

        Parameters
        ----------
        obj : symengine object
        symbols : dict mapping symengine.Symbol to symengine object

        Returns
        -------
        symengine object
        """
        try:
            # Need to do more substitutions to catch symbols that are functions
            # of other symbols
            for iteration in range(_MAX_PARAM_NESTING):
                obj = obj.xreplace(symbols)
                undefs = [x for x in obj.free_symbols if not isinstance(x, v.StateVariable)]
                if len(undefs) == 0:
                    break
        except AttributeError:
            # Can't use xreplace on a float
            pass
        return obj

    def _build_reference_model(self, preserve_ideal=True):
        raise NotImplementedError()

    @property
    def reference_model(self):
        raise NotImplementedError("Endmember reference models do not have a physical meaning for MQMQA models")



""""Common pyparsing grammar patterns."""

from pyparsing import alphas, nums
from pyparsing import Group, OneOrMore, Optional, Regex, Suppress, Word
import re

pos_neg_int_number = Word('+-' + nums).setParseAction(lambda t: [int(t[0])])  # '+3' or '-2' are examples
# matching float w/ regex is ugly but is recommended by pyparsing
regex_after_decimal = r'([0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
float_number = Regex(r'[-+]?([0-9]+\.(?!([0-9]|[eE])))|{0}'.format(regex_after_decimal)) \
    .setParseAction(lambda t: [float(t[0])])

chemical_formula = Group(OneOrMore(Word(alphas, min=1, max=2) + Optional(float_number, default=1.0))) + \
                   Optional(Suppress('/') + pos_neg_int_number, default=0)

reg_symbol = r'([A-z][A-z]?)'
reg_amount = r'([-+]?([0-9]+\.(?!([0-9]|[eE])))|([0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?))?'
reg_charge = r'/([+-]?[0-9]+)'
chem_regex = reg_symbol + reg_amount

def parse_chemical_formula(formula):
    """"""
    matches = re.findall(chem_regex, formula)
    sym_amnts = [(m[0], float(m[1]) if m[1] != '' else 1.0) for m in matches]
    charge = re.search(reg_charge, formula)
    if charge is None:
        charge = 0
    else:
        charge = int(charge.groups()[0])
    return (sym_amnts, charge)



"""
The eqplot module contains functions for general plotting of
the results of equilibrium calculations.
"""
from pycalphad.core.utils import unpack_condition
from pycalphad.plot.utils import phase_legend
import pycalphad.variables as v
from matplotlib import collections as mc
import matplotlib.pyplot as plt
import numpy as np
from collections import OrderedDict

# TODO: support other state variables here or make isinstance elif == v.T or v.P
_plot_labels = {v.T: 'Temperature (K)', v.P: 'Pressure (Pa)'}


def _axis_label(ax_var):
    if isinstance(ax_var, v.MoleFraction):
        return 'X({})'.format(ax_var.species.name)
    elif isinstance(ax_var, v.StateVariable):
        return _plot_labels[ax_var]
    else:
        return ax_var

def _map_coord_to_variable(coord):
    """
    Map a coordinate to a StateVariable object.

    Parameters
    ----------
    coord : str
        Name of coordinate in equilibrium object.

    Returns
    -------
    pycalphad StateVariable
    """
    vals = {'T': v.T, 'P': v.P}
    if coord.startswith('X_'):
        return v.X(coord[2:])
    elif coord in vals:
        return vals[coord]
    else:
        return coord


def eqplot(eq, ax=None, x=None, y=None, z=None, tielines=True, tieline_color=(0, 1, 0, 1), tie_triangle_color=(1, 0, 0, 1), legend_generator=phase_legend, **kwargs):
    """
    Plot the result of an equilibrium calculation.

    The type of plot is controlled by the degrees of freedom in the equilibrium calculation.

    Parameters
    ----------
    eq : xarray.Dataset
        Result of equilibrium calculation.
    ax : matplotlib.Axes
        Default axes used if not specified.
    x : StateVariable, optional
    y : StateVariable, optional
    z : StateVariable, optional
    tielines : bool
        If True, will plot tielines
    tieline_color: color
        A valid matplotlib color, such as a named color string, hex RGB
        string, or a tuple of RGBA components to set the color of the two
        phase region tielines. The default is an RGBA tuple for green:
        (0, 1, 0, 1).
    tie_triangle_color: color
        A valid matplotlib color, such as a named color string, hex RGB
        string, or a tuple of RGBA components to set the color of the two
        phase region tielines. The default is an RGBA tuple for red:
        (1, 0, 0, 1).
    legend_generator : Callable
        A function that will be called with the list of phases and will
        return legend labels and colors for each phase. By default
        pycalphad.plot.utils.phase_legend is used
    kwargs : kwargs
        Passed to `matplotlib.pyplot.scatter`.

    Returns
    -------
    matplotlib AxesSubplot
    """
    conds = OrderedDict([(_map_coord_to_variable(key), unpack_condition(np.asarray(value)))
                         for key, value in sorted(eq.coords.items(), key=str)
                         if (key in ('T', 'P', 'N')) or (key.startswith('X_'))])
    indep_comps = sorted([key for key, value in conds.items() if isinstance(key, v.MoleFraction) and len(value) > 1], key=str)
    indep_pots = [key for key, value in conds.items() if (type(key) is v.StateVariable) and len(value) > 1]

    # determine what the type of plot will be
    if len(indep_comps) == 1 and len(indep_pots) == 1:
        projection = None
    elif len(indep_comps) == 2 and len(indep_pots) == 0:
        projection = 'triangular'
    else:
        raise ValueError('The eqplot projection is not defined and cannot be autodetected. There are {} independent compositions and {} indepedent potentials.'.format(len(indep_comps), len(indep_pots)))
    if z is not None:
        raise NotImplementedError('3D plotting is not yet implemented')
    if ax is None:
        fig, (ax) = plt.subplots(subplot_kw={'projection': projection})

    # Handle cases for different plot types
    if projection is None:
        x = indep_comps[0] if x is None else x
        y = indep_pots[0] if y is None else y
        # plot settings
        ax.set_xlim([np.min(conds[x]) - 1e-2, np.max(conds[x]) + 1e-2])
        ax.set_ylim([np.min(conds[y]), np.max(conds[y])])
    elif projection == 'triangular':
        x = indep_comps[0] if x is None else x
        y = indep_comps[1] if y is None else y
        # Here we adjust the x coordinate of the ylabel.
        # We make it reasonably comparable to the position of the xlabel from the xaxis
        # As the figure size gets very large, the label approaches ~0.55 on the yaxis
        # 0.55*cos(60 deg)=0.275, so that is the xcoord we are approaching.
        ax.yaxis.label.set_va('baseline')
        fig_x_size = ax.figure.get_size_inches()[0]
        y_label_offset = 1 / fig_x_size
        ax.yaxis.set_label_coords(x=(0.275 - y_label_offset), y=0.5)

    # get the active phases and support loading netcdf files from disk
    phases = map(str, sorted(set(np.array(eq.Phase.values.ravel(), dtype='U')) - {''}, key=str))
    comps = map(str, sorted(np.array(eq.coords['component'].values, dtype='U'), key=str))
    eq['component'] = np.array(eq['component'], dtype='U')
    eq['Phase'].values = np.array(eq['Phase'].values, dtype='U')

    # Select all two- and three-phase regions
    three_phase_idx = np.nonzero(np.sum(eq.Phase.values != '', axis=-1, dtype=np.int_) == 3)
    two_phase_idx = np.nonzero(np.sum(eq.Phase.values != '', axis=-1, dtype=np.int_) == 2)

    legend_handles, colorlist = legend_generator(phases)

    # For both two and three phase, cast the tuple of indices to an array and flatten
    # If we found two phase regions:
    if two_phase_idx[0].size > 0:
        found_two_phase = eq.Phase.values[two_phase_idx][..., :2]
        # get tieline endpoint compositions
        two_phase_x = eq.X.sel(component=x.species.name).values[two_phase_idx][..., :2]
        # handle special case for potential
        if isinstance(y, v.MoleFraction):
            two_phase_y = eq.X.sel(component=y.species.name).values[two_phase_idx][..., :2]
        else:
            # it's a StateVariable. This must be True
            two_phase_y = np.take(eq[str(y)].values, two_phase_idx[list(str(i) for i in conds.keys()).index(str(y))])
            # because the above gave us a shape of (n,) instead of (n,2) we are going to create it ourselves
            two_phase_y = np.array([two_phase_y, two_phase_y]).swapaxes(0, 1)

        # plot two phase points
        two_phase_plotcolors = np.array(list(map(lambda x: [colorlist[x[0]], colorlist[x[1]]], found_two_phase)), dtype='U')
        ax.scatter(two_phase_x[..., 0], two_phase_y[..., 0], s=3, c=two_phase_plotcolors[:, 0], edgecolors='None', zorder=2, **kwargs)
        ax.scatter(two_phase_x[..., 1], two_phase_y[..., 1], s=3, c=two_phase_plotcolors[:, 1], edgecolors='None', zorder=2, **kwargs)

        if tielines:
            # construct and plot tielines
            two_phase_tielines = np.array([np.concatenate((two_phase_x[..., 0][..., np.newaxis], two_phase_y[..., 0][..., np.newaxis]), axis=-1),
                                           np.concatenate((two_phase_x[..., 1][..., np.newaxis], two_phase_y[..., 1][..., np.newaxis]), axis=-1)])
            two_phase_tielines = np.rollaxis(two_phase_tielines, 1)
            lc = mc.LineCollection(two_phase_tielines, zorder=1, colors=tieline_color, linewidths=[0.5, 0.5])
            ax.add_collection(lc)

    # If we found three phase regions:
    if (three_phase_idx[0].size > 0) and (len(indep_comps) == 2):
        found_three_phase = eq.Phase.values[three_phase_idx][..., :3]
        # get tieline endpoints
        three_phase_x = eq.X.sel(component=x.species.name).values[three_phase_idx][..., :3]
        three_phase_y = eq.X.sel(component=y.species.name).values[three_phase_idx][..., :3]
        # three phase tielines, these are tie triangles and we always plot them
        three_phase_tielines = np.array([np.concatenate((three_phase_x[..., 0][..., np.newaxis], three_phase_y[..., 0][..., np.newaxis]), axis=-1),
                                         np.concatenate((three_phase_x[..., 1][..., np.newaxis], three_phase_y[..., 1][..., np.newaxis]), axis=-1),
                                         np.concatenate((three_phase_x[..., 2][..., np.newaxis], three_phase_y[..., 2][..., np.newaxis]), axis=-1)])
        three_phase_tielines = np.rollaxis(three_phase_tielines, 1)
        three_lc = mc.LineCollection(three_phase_tielines, zorder=1, colors=tie_triangle_color, linewidths=[0.5, 0.5])
        # plot three phase points and tielines
        three_phase_plotcolors = np.array(list(map(lambda x: [colorlist[x[0]], colorlist[x[1]], colorlist[x[2]]], found_three_phase)), dtype='U')
        ax.scatter(three_phase_x[..., 0], three_phase_y[..., 0], s=3, c=three_phase_plotcolors[:, 0], edgecolors='None', zorder=2, **kwargs)
        ax.scatter(three_phase_x[..., 1], three_phase_y[..., 1], s=3, c=three_phase_plotcolors[:, 1], edgecolors='None', zorder=2, **kwargs)
        ax.scatter(three_phase_x[..., 2], three_phase_y[..., 2], s=3, c=three_phase_plotcolors[:, 2], edgecolors='None', zorder=2, **kwargs)
        ax.add_collection(three_lc)

    # position the phase legend and configure plot
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
    ax.legend(handles=legend_handles, loc='center left', bbox_to_anchor=(1, 0.5))
    ax.tick_params(axis='both', which='major', labelsize=14)
    ax.grid(True)
    plot_title = '-'.join([component.title() for component in sorted(comps) if component != 'VA'])
    ax.set_title(plot_title, fontsize=20)
    ax.set_xlabel(_axis_label(x), labelpad=15, fontsize=20)
    ax.set_ylabel(_axis_label(y), fontsize=20)

    return ax



import numpy as np
import warnings

class BinaryCompset():
    """
    Composition set for representations of a phase equilibria in a two
    component system with a temperature condition.

    Parameters
    ----------
    phase_name : str
        Name of phase
    temperature : float
        Temperature corresponding the the calculation
    indep_comp : str
        Name of the independent component
    composition : float
        Mole fraction of the independent component
    site_fracs : np.ndarray
        Array of floats corresponding to the site fractions.

    Notes
    -----
    In the future, this representation could be phased out if `equilibrium` returned composition sets.
    """
    # tolerances for defining equality between composition sets
    SITE_FRAC_ATOL = 0.001
    TEMPERATURE_ATOL = 0.1

    def __init__(self, phase_name, temperature, indep_comp, composition, site_fracs):
        self.phase_name = phase_name
        self.temperature = temperature
        self.indep_comp = indep_comp
        self.composition = composition
        self.site_fracs = site_fracs

    def __repr__(self,):
        return "<BinaryCompset({0}, T={1:0.3f}, X({2})={3:0.3f})>".format(self.phase_name, self.temperature, self.indep_comp, self.composition)

    def __str__(self,):
        return self.__repr__()

    def __eq__(self, other):
        if hasattr(other, 'phase_name') and hasattr(other, 'site_fracs') and hasattr(other, 'temperature'):
            same_phase = self.phase_name == other.phase_name
            site_frac_close = np.all(np.isclose(self.site_fracs, other.site_fracs, atol=self.__class__.SITE_FRAC_ATOL))
            temp_close = np.isclose(self.temperature, other.temperature, atol=self.__class__.TEMPERATURE_ATOL)
            return same_phase and site_frac_close and temp_close
        else:
            return False

    @classmethod
    def from_dataset_vertices(cls, ds, indep_comp, indep_comp_idx, num_vertex):
        compsets = []
        Phase = ds["Phase"].squeeze()
        T = np.array(ds["T"]).squeeze()
        X = ds["X"].squeeze()
        Y = ds["Y"].squeeze()
        for i in range(num_vertex):
            phase_name = Phase[i]
            if phase_name != '':
                compsets.append(cls(str(phase_name),
                                    T,
                                    indep_comp,
                                    X[i, indep_comp_idx],
                                    Y[i, :],
                                    ))
        return compsets


class CompsetPair():
    """
    Pair of binary composition sets that make up an equilibrium

    Parameters
    ----------
    compsets : list of BinaryCompset

    Attributes
    ----------
    compsets : list of BinaryCompset
        CompSets sorted by composition
    a : BinaryCompset
        Composition set in the pair with the lower composition
    b : BinaryCompset
        Composition set in the pair with the higher composition
    phases : list of str
        List of phase names, sorted by composition
    unique_phases : set of str
        Set of phase names
    compositions : np.ndarray
        Array of compositions, sorted
    temperature : float
        Temperature of the 2D compsets
    same_phase : bool
        Whether the composition sets are for the same phase
    mean_composition : float
        Mean composition of the CompSets
    max_composition : float
        Max composition of the CompSets
    min_composition : float
        Min composition of the CompSets
    """
    def __init__(self, compsets):
        _composition_key_func = lambda c: c.composition
        sorted_compsets = sorted(compsets, key=_composition_key_func)
        phases = [c.phase_name for c in sorted_compsets]
        compositions = np.array([c.composition for c in sorted_compsets])
        a = sorted_compsets[0]
        b = sorted_compsets[1]

        self._orig_compsets = compsets  # with original sorting
        self.compsets = sorted_compsets
        self.phases = phases
        self.unique_phases = frozenset(phases)
        self.a = a
        self.b = b
        self.same_phase = a.phase_name == b.phase_name
        self.compositions = compositions
        self.mean_composition = np.mean(compositions)
        self.min_composition = a.composition
        self.max_composition = b.composition
        if a.temperature == b.temperature:
            self.temperature = a.temperature
        else:
            warnings.warn("Temperatures are different for CompSet objects {}. Assuming that the pair temperature is ".format(sorted_compsets, a.temperature))
            self.temperature = a.temperature

    def xdiscrepancy(self, ignore_phase=False):
        """
        Calculate the composition discrepancy (absolute difference) between this
        composition set and another.

        Parameters
        ----------
        ignore_phase : bool
            If False, unlike phases will give infinite discrepancy. If True, we
            only care about the composition and the real discrepancy will be returned.

        Returns
        -------
        np.float64

        """
        if self.same_phase or ignore_phase:
            return np.abs(self.a.composition - self.b.composition)
        else:
            return np.inf

    def ydiscrepancy(self):
        """
        Calculate the discrepancy (absolute differences) between the site
        fractions of the composition sets as an array of discrepancies.

        Returns
        -------
        Array of np.float64

        Notes
        -----
        The phases must match for this to be meaningful.

        """
        if self.same_phase:
            return np.abs(self.a.site_fracs - self.b.site_fracs)
        else:
            return np.inf

    def ydiscrepancy_max(self):
        """
        Calculate the maximum discrepancy (absolute difference) between the site
        fractions of the composition sets.

        Returns
        -------
        np.float64

        Notes
        -----
        The phases must match for this to be meaningful.

        """
        return np.max(np.abs(self.ydiscrepancy()))

    def Tdiscrepancy(self, other):
        """
        Calculate the temperature discrepancy (absolute difference) between this
        pair of composition sets and another.

        Parameters
        ----------
        other : CompsetPair

        Returns
        -------
        np.float64

        """
        return np.abs(self.temperature - other.temperature)

    def __repr__(self,):
        compset_strs = ", ".join(["{0}:<X({1})={2:0.3f}>".format(c.phase_name, c.indep_comp, c.composition) for c in self.compsets])
        return "CompSet2D(T={0:0.3f}: ({1})".format(self.temperature, compset_strs)

    def __str__(self,):
        return self.__repr__()

    def pairwise_xdiscrepancy(self, other):
        """
        Compute the ordered composition discrepancy between this and another
        pair of CompSet2D objects.

        Parameters
        ----------
        other : CompsetPair

        Returns
        -------
        float
        """
        if self.phases == other.phases:
            return np.abs(self.compositions - other.compositions)
        else:
            return np.full(self.compositions.shape, np.inf)


def get_compsets(eq_dataset, indep_comp=None, indep_comp_index=None):
    """
    Return a CompSet2D object if a pair of composition sets is found in an
    equilibrium dataset. Otherwise return None.

    Parameters
    ----------
    eq_dataset :
    indep_comp :
    indep_comp_index :

    Returns
    -------
    CompsetPair
    """
    if indep_comp is None:
        indep_comp = [c for c in eq_dataset.coords if 'X_' in c][0][2:]
    if indep_comp_index is None:
        indep_comp_index = eq_dataset.component.values.tolist().index(indep_comp)
    extracted_compsets = BinaryCompset.from_dataset_vertices(eq_dataset, indep_comp, indep_comp_index, 3)
    if len(extracted_compsets) == 2:
        return CompsetPair(extracted_compsets)
    else:
        return None


def find_two_phase_region_compsets(hull, temperature, indep_comp, indep_comp_idx, misc_gap_tol=0.1, minimum_composition=None):
    """
    From a 1D convex hull at constant T and P, return the composition sets for
    a two phase region or that have the smallest index composition coordinate

    Parameters
    ----------
    hull : EquilibriumResult
        Equilibrium-like from pycalphad that has a `Phase` Data variable.
    temperature : float
        Temperature that the calculation was performed at
    indep_comp : str
        Name of the independent component
    indep_comp_idx : str
        Index of the independent component in the the sorted pure elements
    misc_gap_tol : float
        If any site fractions are different by at least this amount, the
        composition sets are considered distinct and in a miscibility gap.
    minimum_composition : float
        Minimum composition in the convex hull to search for composition sets

    Returns
    -------
    CompsetPair

    """
    phases = hull.Phase.squeeze()
    compositions = hull.X.squeeze()
    site_fracs = hull.Y.squeeze()
    grid_shape = phases.shape[:-1]
    num_phases = phases.shape[-1]
    it = np.nditer(np.empty(grid_shape), flags=['multi_index'])  # empty grid for indexing
    while not it.finished:
        idx = it.multi_index
        cs = []
        # TODO: assumption of only two phases, seems like the third phase index can have bad points
        # Three phases is probably an error anyways...
        if minimum_composition is not None and np.all(compositions[idx][:, indep_comp_idx][:2] < minimum_composition):
            it.iternext()
            continue
        for i in np.arange(num_phases):
            if str(phases[idx][i]) != '':
                stable_composition_sets = BinaryCompset(str(phases[idx][i]), temperature, indep_comp, compositions[idx][i, indep_comp_idx], site_fracs[idx][i, :])
                cs.append(stable_composition_sets)
        if len(cs) == 2:
            compsets = CompsetPair(cs)
            if len(compsets.unique_phases) == 2:
                return compsets  # found a multiphase region
            else:
                # Same phase, either single phase region or a miscibility gap.
                y_discrep = compsets.ydiscrepancy()
                if np.any(y_discrep[~np.isnan(y_discrep)] > misc_gap_tol):
                    return compsets  # miscibility gap
        it.iternext()
    return None



"""
The equilibrium module defines routines for interacting with
calculated phase equilibria.
"""
import warnings
from collections import OrderedDict
from collections.abc import Mapping
from datetime import datetime
import pycalphad.variables as v
from pycalphad.core.utils import unpack_components, unpack_condition, unpack_phases, filter_phases, instantiate_models, get_state_variables
from pycalphad import calculate
from pycalphad.core.errors import EquilibriumError, ConditionError
from pycalphad.core.starting_point import starting_point
from pycalphad.codegen.callables import build_phase_records
from pycalphad.core.eqsolver import _solve_eq_at_conditions
from pycalphad.core.phase_rec import PhaseRecord
from pycalphad.core.solver import Solver
from pycalphad.core.light_dataset import LightDataset
from pycalphad.model import Model
import numpy as np


def _adjust_conditions(conds):
    "Adjust conditions values to be within the numerical limit of the solver."
    new_conds = OrderedDict()
    minimum_composition = 1e-10
    for key, value in sorted(conds.items(), key=str):
        if key == str(key):
            key = getattr(v, key, key)
        if isinstance(key, v.MoleFraction):
            vals = unpack_condition(value)
            # "Zero" composition is a common pattern. Do not warn for that case.
            if np.any(np.logical_and(np.asarray(vals) < minimum_composition, np.asarray(vals) > 0)):
                warnings.warn(
                    f"Some specified compositions are below the minimum allowed composition of {minimum_composition}.")
            new_conds[key] = [max(val, minimum_composition) for val in vals]
        else:
            new_conds[key] = unpack_condition(value)
    return new_conds


def _eqcalculate(dbf, comps, phases, conditions, output, data=None, per_phase=False, callables=None, model=None,
                 parameters=None, **kwargs):
    """
    WARNING: API/calling convention not finalized.
    Compute the *equilibrium value* of a property.
    This function differs from `calculate` in that it computes
    thermodynamic equilibrium instead of randomly sampling the
    internal degrees of freedom of a phase.
    Because of that, it's slower than `calculate`.
    This plugs in the equilibrium phase and site fractions
    to compute a thermodynamic property defined in a Model.

    Parameters
    ----------
    dbf : Database
        Thermodynamic database containing the relevant parameters.
    comps : list
        Names of components to consider in the calculation.
    phases : list or dict
        Names of phases to consider in the calculation.
    conditions : dict or (list of dict)
        StateVariables and their corresponding value.
    output : str
        Equilibrium model property (e.g., CPM, HM, etc.) to compute.
        This must be defined as an attribute in the Model class of each phase.
    data : Dataset
        Previous result of call to `equilibrium`.
        Should contain the equilibrium configurations at the conditions of interest.
        If the databases are not the same as in the original calculation,
        the results may be meaningless.
    per_phase : bool, optional
        If True, compute and return the property for each phase present.
        If False, return the total system value, weighted by the phase fractions.
    callables : dict
        Callable functions to compute 'output' for each phase.
    model : a dict of phase names to Model
        Model class to use for each phase.
    parameters : dict, optional
        Maps SymEngine Symbol to numbers, for overriding the values of parameters in the Database.
    kwargs
        Passed to `calculate`.

    Returns
    -------
    Dataset of property as a function of equilibrium conditions
    """
    if data is None:
        raise ValueError('Required kwarg "data" is not specified')
    if model is None:
        raise ValueError('Required kwarg "model" is not specified')
    active_phases = unpack_phases(phases)
    conds = _adjust_conditions(conditions)
    indep_vars = ['N', 'P', 'T']
    # TODO: Rewrite this to use the coord dict from 'data'
    str_conds = OrderedDict((str(key), value) for key, value in conds.items())
    indep_vals = list([float(x) for x in np.atleast_1d(val)]
                      for key, val in str_conds.items() if key in indep_vars)
    coord_dict = str_conds.copy()
    components = [x for x in sorted(comps)]
    desired_active_pure_elements = [list(x.constituents.keys()) for x in components]
    desired_active_pure_elements = [el.upper() for constituents in desired_active_pure_elements for el in constituents]
    pure_elements = sorted(set([x for x in desired_active_pure_elements if x != 'VA']))
    coord_dict['vertex'] = np.arange(len(pure_elements) + 1)  # +1 is to accommodate the degenerate degree of freedom at the invariant reactions
    grid_shape = np.meshgrid(*coord_dict.values(),
                             indexing='ij', sparse=False)[0].shape
    prop_shape = grid_shape
    prop_dims = list(str_conds.keys()) + ['vertex']

    result = LightDataset({output: (prop_dims, np.full(prop_shape, np.nan))}, coords=coord_dict)
    # For each phase select all conditions where that phase exists
    # Perform the appropriate calculation and then write the result back
    for phase in active_phases:
        dof = len(model[phase].site_fractions)
        current_phase_indices = (data.Phase == phase)
        if ~np.any(current_phase_indices):
            continue
        points = data.Y[np.nonzero(current_phase_indices)][..., :dof]
        statevar_indices = np.nonzero(current_phase_indices)[:len(indep_vals)]
        statevars = {key: np.take(np.asarray(vals), idx)
                     for key, vals, idx in zip(indep_vars, indep_vals, statevar_indices)}
        statevars.update(kwargs)
        if statevars.get('mode', None) is None:
            statevars['mode'] = 'numpy'
        calcres = calculate(dbf, comps, [phase], output=output, points=points, broadcast=False,
                            callables=callables, parameters=parameters, model=model, **statevars)
        result[output][np.nonzero(current_phase_indices)] = calcres[output].values
    if not per_phase:
        out = np.nansum(result[output] * data['NP'], axis=-1)
        dv_output = result.data_vars[output]
        result.remove(output)
        # remove the vertex coordinate because we summed over it
        result.add_variable(output, dv_output[0][:-1], out)
    else:
        dv_phase = data.data_vars['Phase']
        dv_np = data.data_vars['NP']
        result.add_variable('Phase', dv_phase[0], dv_phase[1])
        result.add_variable('NP', dv_np[0], dv_np[1])
    return result


def equilibrium(dbf, comps, phases, conditions, output=None, model=None,
                verbose=False, broadcast=True, calc_opts=None, to_xarray=True,
                scheduler='sync', parameters=None, solver=None, callables=None,
                phase_records=None, **kwargs):
    """
    Calculate the equilibrium state of a system containing the specified
    components and phases, under the specified conditions.

    Parameters
    ----------
    dbf : Database
        Thermodynamic database containing the relevant parameters.
    comps : list
        Names of components to consider in the calculation.
    phases : list or dict
        Names of phases to consider in the calculation.
    conditions : dict or (list of dict)
        StateVariables and their corresponding value.
    output : str or list of str, optional
        Additional equilibrium model properties (e.g., CPM, HM, etc.) to compute.
        These must be defined as attributes in the Model class of each phase.
    model : Model, a dict of phase names to Model, or a seq of both, optional
        Model class to use for each phase.
    verbose : bool, optional
        Print details of calculations. Useful for debugging.
    broadcast : bool
        If True, broadcast conditions against each other. This will compute all combinations.
        If False, each condition should be an equal-length list (or single-valued).
        Disabling broadcasting is useful for calculating equilibrium at selected conditions,
        when those conditions don't comprise a grid.
    calc_opts : dict, optional
        Keyword arguments to pass to `calculate`, the energy/property calculation routine.
    to_xarray : bool
        Whether to return an xarray Dataset (True, default) or an EquilibriumResult.
    scheduler : Dask scheduler, optional
        Job scheduler for performing the computation.
        If None, return a Dask graph of the computation instead of actually doing it.
    parameters : dict, optional
        Maps SymEngine Symbol to numbers, for overriding the values of parameters in the Database.
    solver : pycalphad.core.solver.SolverBase
        Instance of a solver that is used to calculate local equilibria.
        Defaults to a pycalphad.core.solver.Solver.
    callables : dict, optional
        Pre-computed callable functions for equilibrium calculation.
    phase_records : Optional[Mapping[str, PhaseRecord]]
        Mapping of phase names to PhaseRecord objects with `'GM'` output. Must include
        all active phases. The `model` argument must be a mapping of phase names to
        instances of Model objects.

    Returns
    -------
    Structured equilibrium calculation, or Dask graph if scheduler=None.

    Examples
    --------
    None yet.
    """
    if not broadcast:
        raise NotImplementedError('Broadcasting cannot yet be disabled')
    comps = sorted(unpack_components(dbf, comps))
    phases = unpack_phases(phases) or sorted(dbf.phases.keys())
    list_of_possible_phases = filter_phases(dbf, comps)
    if len(list_of_possible_phases) == 0:
        raise ConditionError('There are no phases in the Database that can be active with components {0}'.format(comps))
    active_phases = filter_phases(dbf, comps, phases)
    if len(active_phases) == 0:
        raise ConditionError('None of the passed phases ({0}) are active. List of possible phases: {1}.'.format(phases, list_of_possible_phases))
    if isinstance(comps, (str, v.Species)):
        comps = [comps]
    if len(set(comps) - set(dbf.species)) > 0:
        raise EquilibriumError('Components not found in database: {}'
                               .format(','.join([c.name for c in (set(comps) - set(dbf.species))])))
    calc_opts = calc_opts if calc_opts is not None else dict()
    solver = solver if solver is not None else Solver(verbose=verbose)
    parameters = parameters if parameters is not None else dict()
    if isinstance(parameters, dict):
        parameters = OrderedDict(sorted(parameters.items(), key=str))
    # Temporary solution until constraint system improves
    if conditions.get(v.N) is None:
        conditions[v.N] = 1
    if np.any(np.array(conditions[v.N]) != 1):
        raise ConditionError('N!=1 is not yet supported, got N={}'.format(conditions[v.N]))
    # Modify conditions values to be within numerical limits, e.g., X(AL)=0
    # Also wrap single-valued conditions with lists
    conds = _adjust_conditions(conditions)

    for cond in conds.keys():
        if isinstance(cond, (v.MoleFraction, v.ChemicalPotential)) and cond.species not in comps:
            raise ConditionError('{} refers to non-existent component'.format(cond))
    str_conds = OrderedDict((str(key), value) for key, value in conds.items())
    components = [x for x in sorted(comps)]
    desired_active_pure_elements = [list(x.constituents.keys()) for x in components]
    desired_active_pure_elements = [el.upper() for constituents in desired_active_pure_elements for el in constituents]
    pure_elements = sorted(set([x for x in desired_active_pure_elements if x != 'VA']))
    if verbose:
        print('Components:', ' '.join([str(x) for x in comps]))
        print('Phases:', end=' ')
    output = output if output is not None else 'GM'
    output = output if isinstance(output, (list, tuple, set)) else [output]
    output = set(output)
    output |= {'GM'}
    output = sorted(output)
    if phase_records is None:
        models = instantiate_models(dbf, comps, active_phases, model=model, parameters=parameters)
        phase_records = build_phase_records(dbf, comps, active_phases, conds, models,
                                            output='GM', callables=callables,
                                            parameters=parameters, verbose=verbose,
                                            build_gradients=True, build_hessians=True)
    else:
        # phase_records were provided, instantiated models must also be provided by the caller
        models = model
        if not isinstance(models, Mapping):
            raise ValueError("A dictionary of instantiated models must be passed to `equilibrium` with the `model` argument if the `phase_records` argument is used.")
        active_phases_without_models = [name for name in active_phases if not isinstance(models.get(name), Model)]
        active_phases_without_phase_records = [name for name in active_phases if not isinstance(phase_records.get(name), PhaseRecord)]
        if len(active_phases_without_phase_records) > 0:
            raise ValueError(f"phase_records must contain a PhaseRecord instance for every active phase. Missing PhaseRecord objects for {sorted(active_phases_without_phase_records)}")
        if len(active_phases_without_models) > 0:
            raise ValueError(f"model must contain a Model instance for every active phase. Missing Model objects for {sorted(active_phases_without_models)}")

    if verbose:
        print('[done]', end='\n')

    state_variables = sorted(get_state_variables(models=models, conds=conds), key=str)

    # 'calculate' accepts conditions through its keyword arguments
    grid_opts = calc_opts.copy()
    statevar_strings = [str(x) for x in state_variables]
    grid_opts.update({key: value for key, value in str_conds.items() if key in statevar_strings})

    if 'pdens' not in grid_opts:
        grid_opts['pdens'] = 60
    grid = calculate(dbf, comps, active_phases, model=models, fake_points=True,
                     phase_records=phase_records, output='GM', parameters=parameters,
                     to_xarray=False, **grid_opts)
    coord_dict = str_conds.copy()
    coord_dict['vertex'] = np.arange(len(pure_elements) + 1)  # +1 is to accommodate the degenerate degree of freedom at the invariant reactions
    coord_dict['component'] = pure_elements
    properties = starting_point(conds, state_variables, phase_records, grid)
    properties = _solve_eq_at_conditions(properties, phase_records, grid,
                                         list(str_conds.keys()), state_variables,
                                         verbose, solver=solver)

    # Compute equilibrium values of any additional user-specified properties
    # We already computed these properties so don't recompute them
    output = sorted(set(output) - {'GM', 'MU'})
    for out in output:
        if (out is None) or (len(out) == 0):
            continue
        # TODO: How do we know if a specified property should be per_phase or not?
        # For now, we make a best guess
        if (out == 'degree_of_ordering') or (out == 'DOO'):
            per_phase = True
        else:
            per_phase = False
        eqcal = _eqcalculate(dbf, comps, active_phases, conditions, out,
                             data=properties, per_phase=per_phase, model=models,
                             callables=callables, parameters=parameters, **calc_opts)
        properties = properties.merge(eqcal, inplace=True, compat='equals')
    if to_xarray:
        properties = properties.get_dataset()
    properties.attrs['created'] = datetime.utcnow().isoformat()
    if len(kwargs) > 0:
        warnings.warn('The following equilibrium keyword arguments were passed, but unused:\n{}'.format(kwargs))
    return properties
